# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import math

from thumbor.ext.filters import _convolution
from thumbor.filters import BaseFilter, filter_method

MAX_RADIUS = 150


def generate_1d_matrix(sigma, radius):
    matrix_size = (radius * 2) + 1
    matrix = []
    two_sigma_squared = float(2 * sigma * sigma)
    for column in range(matrix_size):
        adj_x = column - radius
        exp = math.e ** -(((adj_x * adj_x)) / two_sigma_squared)
        matrix.append(exp / math.sqrt(two_sigma_squared * math.pi))
    return tuple(matrix), matrix_size


def apply_blur(mode, data, size, radius, sigma=0):
    if sigma == 0:
        sigma = radius
    radius = min(radius, MAX_RADIUS)
    matrix, matrix_size = generate_1d_matrix(sigma, radius)
    data = _convolution.apply(
        mode, data, size[0], size[1], matrix, matrix_size, True
    )
    return _convolution.apply(mode, data, size[0], size[1], matrix, 1, True)


class Filter(BaseFilter):
    """
    Usage: /filters:blur(<radius> [, <sigma>])
    Examples of use:
        /filters:blur(1)/
        /filters:blur(4)/
        /filters:blur(4, 2)/
    """

    @filter_method(BaseFilter.PositiveNonZeroNumber, BaseFilter.DecimalNumber)
    async def blur(self, radius, sigma=0):
        mode, imgdata = self.engine.image_data_as_rgb()
        imgdata = apply_blur(mode, imgdata, self.engine.size, radius, sigma)
        self.engine.set_image_data(imgdata)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import re
import tempfile
from shutil import rmtree

from preggy import expect
from tornado.testing import gen_test

from tests.base import TestCase
from tests.fixtures.images import (
    TOO_SMALL_IMAGE_PATH,
    VALID_IMAGE_PATH,
    too_heavy_image,
    too_small_image,
    valid_image,
)
from thumbor.config import Config
from thumbor.context import Context
from thumbor.importer import Importer


class UploadTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.root_path = tempfile.mkdtemp()
        cls.base_uri = "/image"

    @classmethod
    def tearDownClass(cls):
        rmtree(cls.root_path)

    @property
    def upload_storage(self):
        return self.context.modules.upload_photo_storage

    def get_path_from_location(self, location):
        return "/".join(location.lstrip("/").rstrip("/").split("/")[1:-1])

    def get_context(self):
        self.default_filename = "image"

        cfg = Config()
        cfg.UPLOAD_ENABLED = True
        cfg.UPLOAD_PHOTO_STORAGE = "thumbor.storages.file_storage"
        cfg.FILE_STORAGE_ROOT_PATH = self.root_path
        cfg.UPLOAD_DELETE_ALLOWED = False
        cfg.UPLOAD_PUT_ALLOWED = False
        cfg.UPLOAD_DEFAULT_FILENAME = self.default_filename

        importer = Importer(cfg)
        importer.import_modules()

        return Context(None, cfg, importer)


class UploadAPINewFileTestCase(UploadTestCase):
    @gen_test
    async def test_can_post_image_with_content_type(self):
        filename = "new_image_with_a_filename.jpg"
        response = await self.async_post(
            self.base_uri,
            {"Content-Type": "image/jpeg", "Slug": filename},
            valid_image(),
        )

        expect(response.code).to_equal(201)

        expect(response.headers).to_include("Location")
        expect(response.headers["Location"]).to_match(
            self.base_uri + r"/[^\/]{32}/" + filename
        )

        expected_path = self.get_path_from_location(
            response.headers["Location"]
        )
        expected_path = self.upload_storage.path_on_filesystem(expected_path)
        expect(expected_path).to_exist()
        expect(expected_path).to_be_the_same_as(VALID_IMAGE_PATH)

    @gen_test
    async def test_can_post_image_with_charset(self):
        filename = self.default_filename + ".jpg"
        response = await self.async_post(
            self.base_uri,
            {"Content-Type": "image/jpeg;charset=UTF-8"},
            valid_image(),
        )

        expect(response.code).to_equal(201)

        expect(response.headers).to_include("Location")
        expect(response.headers["Location"]).to_match(
            self.base_uri + r"/[^\/]{32}/" + filename
        )

        expected_path = self.get_path_from_location(
            response.headers["Location"]
        )
        expected_path = self.upload_storage.path_on_filesystem(expected_path)
        expect(expected_path).to_exist()
        expect(expected_path).to_be_the_same_as(VALID_IMAGE_PATH)

    @gen_test
    async def test_can_post_image_with_unknown_charset(self):
        filename = self.default_filename + ".jpg"
        response = await self.async_post(
            self.base_uri,
            {"Content-Type": "image/thisIsAUnknwonOrBadlyFormedCHarset"},
            valid_image(),
        )

        expect(response.code).to_equal(201)

        expect(response.headers).to_include("Location")
        expect(response.headers["Location"]).to_match(
            self.base_uri + r"/[^\/]{32}/" + filename
        )

        expected_path = self.get_path_from_location(
            response.headers["Location"]
        )
        expected_path = self.upload_storage.path_on_filesystem(expected_path)
        expect(expected_path).to_exist()
        expect(expected_path).to_be_the_same_as(VALID_IMAGE_PATH)

    @gen_test
    async def test_can_post_image_without_filename(self):
        filename = self.default_filename + ".jpg"
        response = await self.async_post(
            self.base_uri, {"Content-Type": "image/jpeg"}, valid_image()
        )

        expect(response.code).to_equal(201)

        expect(response.headers).to_include("Location")
        expect(response.headers["Location"]).to_match(
            self.base_uri + r"/[^\/]{32}/" + filename
        )

        expected_path = self.get_path_from_location(
            response.headers["Location"]
        )
        expected_path = self.upload_storage.path_on_filesystem(expected_path)
        expect(expected_path).to_exist()
        expect(expected_path).to_be_the_same_as(VALID_IMAGE_PATH)

    @gen_test
    async def test_can_post_from_html_form(self):
        filename = "crocodile2.jpg"
        image = ("media", filename, valid_image())
        response = await self.async_post_files(
            self.base_uri, {"Slug": b"another_filename.jpg"}, (image,)
        )

        expect(response.code).to_equal(201)

        expect(response.headers).to_include("Location")
        expect(response.headers["Location"]).to_match(
            self.base_uri + r"/[^\/]{32}/" + filename
        )

        expected_path = self.get_path_from_location(
            response.headers["Location"]
        )
        expected_path = self.upload_storage.path_on_filesystem(expected_path)
        expect(expected_path).to_exist()
        expect(expected_path).to_be_the_same_as(VALID_IMAGE_PATH)


class UploadAPIUpdateFileTestCase(UploadTestCase):
    def get_context(self):
        self.default_filename = "image"

        cfg = Config()
        cfg.UPLOAD_ENABLED = True
        cfg.UPLOAD_PHOTO_STORAGE = "thumbor.storages.file_storage"
        cfg.FILE_STORAGE_ROOT_PATH = self.root_path
        cfg.UPLOAD_DELETE_ALLOWED = False
        cfg.UPLOAD_PUT_ALLOWED = True
        cfg.UPLOAD_DEFAULT_FILENAME = self.default_filename

        importer = Importer(cfg)
        importer.import_modules()

        return Context(None, cfg, importer)

    @gen_test
    async def test_can_modify_existing_image(self):
        filename = self.default_filename + ".jpg"
        response = await self.async_post(
            self.base_uri, {"Content-Type": "image/jpeg"}, valid_image()
        )
        location = response.headers["Location"]
        response = await self.async_put(
            location, {"Content-Type": "image/jpeg"}, too_small_image()
        )

        expect(response.code).to_equal(204)

        id_should_exist = (
            re.compile(self.base_uri + r"/([^\/]{32})/" + filename)
            .search(location)
            .group(1)
        )
        expected_path = self.upload_storage.path_on_filesystem(id_should_exist)
        expect(expected_path).to_exist()
        expect(expected_path).to_be_the_same_as(TOO_SMALL_IMAGE_PATH)

        id_shouldnt_exist = (
            re.compile(self.base_uri + r"/(.*)").search(location).group(1)
        )
        expected_path = self.upload_storage.path_on_filesystem(
            id_shouldnt_exist
        )
        expect(expected_path).not_to_exist()


class UploadAPIUpdateSmallIdFileTestCase(UploadTestCase):
    def get_context(self):
        self.default_filename = "image"

        cfg = Config()
        cfg.UPLOAD_ENABLED = True
        cfg.UPLOAD_PHOTO_STORAGE = "thumbor.storages.file_storage"
        cfg.FILE_STORAGE_ROOT_PATH = self.root_path
        cfg.UPLOAD_DELETE_ALLOWED = False
        cfg.UPLOAD_PUT_ALLOWED = True
        cfg.UPLOAD_DEFAULT_FILENAME = self.default_filename
        cfg.MAX_ID_LENGTH = 36

        importer = Importer(cfg)
        importer.import_modules()

        return Context(None, cfg, importer)

    @gen_test
    async def test_cant_get_truncated_id_when_stored_with_large_id(self):
        image_id = "e5bcf126-791b-4375-9f73-925ab8b9fb5f"
        path = f"/image/{image_id}"

        response = await self.async_put(
            path, {"Content-Type": "image/jpeg"}, valid_image()
        )
        expect(response.code).to_equal(204)

        response = await self.async_get(
            path[: 7 + 32], {"Accept": "image/jpeg"}
        )
        expect(response.code).to_equal(404)

    @gen_test
    async def test_can_get_actual_id_when_stored_with_large_id(self):
        path = "/image/e5bcf126-791b-4375-9f73-925ab8b9fb5g"

        await self.async_put(
            path, {"Content-Type": "image/jpeg"}, valid_image()
        )
        response = await self.async_get(
            path + "123456", {"Accept": "image/jpeg"}
        )

        expect(response.code).to_equal(200)

        expect(response.body).to_be_similar_to(valid_image())


class UploadAPIDeleteTestCase(UploadTestCase):
    def get_context(self):
        self.default_filename = "image"

        cfg = Config()
        cfg.UPLOAD_ENABLED = True
        cfg.UPLOAD_PHOTO_STORAGE = "thumbor.storages.file_storage"
        cfg.FILE_STORAGE_ROOT_PATH = self.root_path
        cfg.UPLOAD_DELETE_ALLOWED = True
        cfg.UPLOAD_PUT_ALLOWED = False
        cfg.UPLOAD_DEFAULT_FILENAME = self.default_filename

        importer = Importer(cfg)
        importer.import_modules()

        return Context(None, cfg, importer)

    @gen_test
    async def test_can_delete_existing_image(self):
        filename = self.default_filename + ".jpg"
        response = await self.async_post(
            self.base_uri, {"Content-Type": "image/jpeg"}, valid_image()
        )
        expect(response.code).to_equal(201)

        location = response.headers["Location"]
        image_id = (
            re.compile(self.base_uri + r"/([^\/]{32})/" + filename)
            .search(location)
            .group(1)
        )
        image_location = self.upload_storage.path_on_filesystem(image_id)
        expect(image_location).to_exist()

        response = await self.async_delete(location, {})
        expect(response.code).to_equal(204)

        expect(image_location).not_to_exist()

    @gen_test
    async def test_deleting_unknown_image_returns_not_found(self):
        uri = self.base_uri + "/an/unknown/image"
        response = await self.async_delete(uri, {})
        expect(response.code).to_equal(404)


class UploadAPIRetrieveTestCase(UploadTestCase):
    def get_context(self):
        self.default_filename = "image"

        cfg = Config()
        cfg.UPLOAD_ENABLED = True
        cfg.UPLOAD_PHOTO_STORAGE = "thumbor.storages.file_storage"
        cfg.FILE_STORAGE_ROOT_PATH = self.root_path
        cfg.UPLOAD_DELETE_ALLOWED = True
        cfg.UPLOAD_PUT_ALLOWED = False
        cfg.UPLOAD_DEFAULT_FILENAME = self.default_filename

        importer = Importer(cfg)
        importer.import_modules()

        return Context(None, cfg, importer)

    @gen_test
    async def test_can_retrieve_existing_image(self):
        response = await self.async_post(
            self.base_uri, {"Content-Type": "image/jpeg"}, valid_image()
        )
        expect(response.code).to_equal(201)

        location = response.headers["Location"]
        response = await self.async_get(location, {"Accept": "image/jpeg"})
        expect(response.code).to_equal(200)
        expect(response.body).to_be_similar_to(valid_image())
        expect(response.headers["Content-Type"]).to_equal("image/jpeg")

    @gen_test
    async def test_retrieving_unknown_image_returns_not_found(self):
        uri = self.base_uri + "/an/unknown/image"
        response = await self.async_get(uri, {"Accept": "image/jpeg"})
        expect(response.code).to_equal(404)


class UploadAPIValidationTestCase(UploadTestCase):
    """
    Validation :
        - Invalid image
        - Size constraints
        - Weight constraints
    """

    def get_context(self):
        self.default_filename = "image"

        cfg = Config()
        cfg.UPLOAD_ENABLED = True
        cfg.UPLOAD_PUT_ALLOWED = True
        cfg.UPLOAD_PHOTO_STORAGE = "thumbor.storages.file_storage"
        cfg.FILE_STORAGE_ROOT_PATH = self.root_path
        cfg.UPLOAD_DEFAULT_FILENAME = self.default_filename
        cfg.MIN_WIDTH = 40
        cfg.MIN_HEIGHT = 40
        cfg.UPLOAD_MAX_SIZE = 72000

        importer = Importer(cfg)
        importer.import_modules()
        return Context(None, cfg, importer)

    @gen_test
    async def test_posting_invalid_image_fails(self):
        response = await self.async_post(
            self.base_uri, {"Content-Type": "image/jpeg"}, "invalid image"
        )
        expect(response.code).to_equal(415)

    @gen_test
    async def test_posting_invalid_image_through_html_form_fails(self):
        image = ("media", "crocodile9999.jpg", b"invalid image")
        response = await self.async_post_files(self.base_uri, {}, (image,))
        expect(response.code).to_equal(415)

    @gen_test
    async def test_modifying_existing_image_to_invalid_image(self):
        response = await self.async_post(
            self.base_uri, {"Content-Type": "image/jpeg"}, valid_image()
        )
        expect(response.code).to_equal(201)
        location = response.headers["Location"]

        response = await self.async_put(
            location, {"Content-Type": "image/jpeg"}, "invalid image"
        )
        expect(response.code).to_equal(415)

        expected_path = self.get_path_from_location(location)
        expected_path = self.upload_storage.path_on_filesystem(expected_path)
        expect(expected_path).to_exist()
        expect(expected_path).to_be_the_same_as(VALID_IMAGE_PATH)

    @gen_test
    async def test_posting_a_too_small_image_fails(self):
        response = await self.async_post(
            self.base_uri, {"Content-Type": "image/jpeg"}, too_small_image()
        )
        expect(response.code).to_equal(412)

    @gen_test
    async def test_posting_a_too_small_image_from_html_form_fails(self):
        image = ("media", "crocodile9999.jpg", too_small_image())
        response = await self.async_post_files(self.base_uri, {}, (image,))
        expect(response.code).to_equal(412)

    @gen_test
    async def test_modifying_existing_image_to_small_image(self):
        response = await self.async_post(
            self.base_uri, {"Content-Type": "image/jpeg"}, valid_image()
        )
        expect(response.code).to_equal(201)

        location = response.headers["Location"]
        response = await self.async_put(
            location, {"Content-Type": "image/jpeg"}, too_small_image()
        )
        expect(response.code).to_equal(412)

        expected_path = self.get_path_from_location(location)
        expected_path = self.upload_storage.path_on_filesystem(expected_path)
        expect(expected_path).to_exist()
        expect(expected_path).to_be_the_same_as(VALID_IMAGE_PATH)

    @gen_test
    async def test_posting_an_image_too_heavy_fails(self):
        response = await self.async_post(
            self.base_uri, {"Content-Type": "image/jpeg"}, too_heavy_image()
        )
        expect(response.code).to_equal(412)

    @gen_test
    async def test_posting_an_image_too_heavy_through_an_html_form_fails(self):
        image = ("media", "oversized9999.jpg", too_heavy_image())
        response = await self.async_post_files(self.base_uri, {}, (image,))
        expect(response.code).to_equal(412)

    @gen_test
    async def test_modifying_existing_image_to_heavy_image_fails(self):
        response = await self.async_post(
            self.base_uri, {"Content-Type": "image/jpeg"}, valid_image()
        )
        location = response.headers["Location"]

        response = await self.async_put(
            location, {"Content-Type": "image/jpeg"}, too_heavy_image()
        )
        expect(response.code).to_equal(412)

        expected_path = self.get_path_from_location(location)
        expected_path = self.upload_storage.path_on_filesystem(expected_path)
        expect(expected_path).to_exist()
        expect(expected_path).to_be_the_same_as(VALID_IMAGE_PATH)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import json
import logging
import logging.handlers
import operator
import re
import time

from thumbor import __version__


class ErrorHandler:
    def __init__(self, config):
        if not config.ERROR_FILE_LOGGER:
            raise RuntimeError(
                "If you set USE_CUSTOM_ERROR_HANDLING to True, and you are "
                "using thumbor_file_logger.logger, then you must specify the file path "
                "to log to with the ERROR_FILE_LOGGER configuration."
            )
        if config.ERROR_FILE_NAME_USE_CONTEXT and not re.search(
            r"^(\w+\.)?\w+$", config.ERROR_FILE_NAME_USE_CONTEXT
        ):
            raise RuntimeError(
                r"ERROR_FILE_NAME_USE_CONTEXT config must refer an attribute of context "
                f"object and be form of ^(\\w+.)?\\w+$ : {config.ERROR_FILE_NAME_USE_CONTEXT}"
            )
        self.file_name = config.ERROR_FILE_LOGGER
        if config.ERROR_FILE_NAME_USE_CONTEXT:
            self.use_context = config.ERROR_FILE_NAME_USE_CONTEXT
        else:
            self.use_context = None
        self.logger = None

    def handle_error(self, context, handler, exception):
        # create log file if not existing
        if not self.logger:
            if self.use_context:
                obj = operator.attrgetter(self.use_context)(context)
                file = self.file_name % obj
            else:
                file = self.file_name

            self.logger = logging.getLogger("error_handler")
            self.logger.setLevel(logging.ERROR)
            self.logger.addHandler(logging.handlers.WatchedFileHandler(file))

        req = handler.request
        extra = {"thumbor-version": __version__, "timestamp": time.time()}
        extra.update({"Headers": req.headers})
        cookies_header = extra.get("Headers", {}).get("Cookie", {})
        if isinstance(cookies_header, str):
            cookies = {}
            for cookie in cookies_header.split(";"):
                if not cookie:
                    continue
                values = cookie.strip().split("=")
                key, val = values[0], "".join(values[1:])
                cookies[key] = val
        else:
            cookies = cookies_header
        extra["Headers"]["Cookie"] = cookies

        data = {
            "Http": {
                "url": req.full_url(),
                "method": req.method,
                "data": req.arguments,
                "body": req.body,
                "query_string": req.query,
            },
            "interfaces.User": {"ip": req.remote_ip},
            "exception": str(exception),
            "extra": extra,
        }

        self.logger.error(json.dumps(data))



import os.path
from shutil import which

from tornado.testing import AsyncHTTPTestCase

from thumbor.app import ThumborServiceApp
from thumbor.config import Config
from thumbor.context import Context, ServerParameters
from thumbor.importer import Importer


class EngineCase(AsyncHTTPTestCase):
    def get_app(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        server_params = ServerParameters(None, None, None, None, None, None)
        server_params.gifsicle_path = which("gifsicle")

        cfg.DETECTORS = [
            "thumbor.detectors.face_detector",
            "thumbor.detectors.profile_detector",
            "thumbor.detectors.glasses_detector",
            "thumbor.detectors.feature_detector",
        ]
        cfg.STORAGE = "thumbor.storages.no_storage"
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = os.path.join(
            os.path.dirname(__file__), "imgs"
        )
        cfg.ENGINE = getattr(self, "engine", None)
        cfg.USE_GIFSICLE_ENGINE = True
        cfg.FFMPEG_PATH = which("ffmpeg")
        cfg.ENGINE_THREADPOOL_SIZE = 10
        cfg.OPTIMIZERS = [
            "thumbor.optimizers.gifv",
        ]
        if not cfg.ENGINE:
            return None

        importer = Importer(cfg)
        importer.import_modules()
        ctx = Context(server_params, cfg, importer)
        application = ThumborServiceApp(ctx)

        return application



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com


from preggy import expect
from tornado.testing import gen_test

from tests.base import FilterTestCase


class BackgroundColorFilterTestCase(FilterTestCase):
    def config_context(self, context):
        context.request.fit_in = True
        context.request.width = 300
        context.request.height = 300

    @gen_test
    async def test_background_color_filter_with_fixed_color(self):
        image = await self.get_filtered(
            "PNG_transparency_demonstration_1.png",
            "thumbor.filters.background_color",
            "background_color(blue)",
            config_context=self.config_context,
        )

        expected = self.get_fixture(
            "PNG_transparency_demonstration_1_blue.png"
        )

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.97)

    @gen_test
    async def test_background_color_filter_with_auto_color(self):
        image = await self.get_filtered(
            "PNG_transparency_demonstration_1.png",
            "thumbor.filters.background_color",
            "background_color(auto)",
            config_context=self.config_context,
        )

        expected = self.get_fixture(
            "PNG_transparency_demonstration_1_auto.png"
        )

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.97)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from typing import Any, cast

from thumbor.handler_lists import HandlerList
from thumbor.handlers.image_resource import ImageResourceHandler
from thumbor.handlers.upload import ImageUploadHandler


def get_handlers(context: Any) -> HandlerList:
    is_upload_enabled = cast(bool, context.config.UPLOAD_ENABLED)
    if not is_upload_enabled:
        return []

    return [
        (r"/image", ImageUploadHandler, {"context": context}),
        (r"/image/(.*)", ImageResourceHandler, {"context": context}),
    ]



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from concurrent.futures import ThreadPoolExecutor

from tornado.ioloop import IOLoop

from thumbor.utils import logger


class ThreadPool:
    @classmethod
    def instance(cls, size):
        """
        Cache threadpool since context is
        recreated for each request
        """
        if not getattr(cls, "_instance", None):
            cls._instance = {}
        if size not in cls._instance:
            cls._instance[size] = ThreadPool(size)
        return cls._instance[size]

    @classmethod
    def reset(cls):
        if cls._instance is not None:
            for instance in cls._instance.values():
                instance.cleanup()
        cls._instance = None

    def __init__(self, thread_pool_size):
        if thread_pool_size:
            self.pool = ThreadPoolExecutor(thread_pool_size)
        else:
            self.pool = None

    async def _execute_in_foreground(self, operation, *args):
        return operation(*args)

    async def _execute_in_pool(self, operation, *args):
        loop = IOLoop.current()
        return await loop.run_in_executor(self.pool, operation, *args)

    async def queue(self, operation, *args):
        if not self.pool:
            return await self._execute_in_foreground(operation, *args)

        return await self._execute_in_pool(operation, *args)

    def cleanup(self):
        if self.pool:
            logger.info("Shutting down threads")
            self.pool.shutdown()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from unittest import TestCase

import thumbor_plugins.optimizers.gifv

import thumbor.optimizers.gifv


class GifvOptimizerTest(TestCase):
    def test_should_be_subclass_of_thumbor_plugins(self):
        self.assertTrue(
            issubclass(
                thumbor.optimizers.gifv.Optimizer,
                thumbor_plugins.optimizers.gifv.Optimizer,
            )
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import glob
import logging
import os

from setuptools import Extension, setup

from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.mkd").read_text()

try:
    from thumbor import __version__
except ImportError:
    __version__ = "0.0.0"

TESTS_REQUIREMENTS = [
    "coverage==6.*,>=6.3.2",
    "flake8==3.*,>=3.7.9",
    "isort==4.*,>=4.3.21",
    "pre-commit==2.*,>=2.17.0",
    "preggy==1.*,>=1.4.4",
    "pylint==2.*,>=2.4.4",
    "pyssim==0.*,>=0.4.0",
    "pytest>=6.2.5",
    "pytest-asyncio==0.*,>=0.10.0",
    "pytest-cov==3.*,>=3.0.0",
    "pytest-tldr==0.*,>=0.2.1",
    "pytest-xdist==2.*,>=2.4.0",
    "redis==4.*,>=4.2.2",
    "remotecv>=2.3.0",
    "sentry-sdk==0.*,>=0.14.1",
    "yanc==0.*,>=0.3.3",
]

OPENCV_REQUIREMENTS = [
    "opencv-python-headless==4.*,>=4.2.0",
    "numpy==1.*,<1.24.0",
]

EXTRA_LIBS_REQUIREMENTS = [
    # Going to update in a proper commit
    "cairosvg>=2.5.2",
    "pycurl==7.*,>=7.43.0",
    "pillow-avif-plugin==1.*,>=1.2.2",
    "pillow-heif>=0.7.0",
]

ALL_REQUIREMENTS = OPENCV_REQUIREMENTS + EXTRA_LIBS_REQUIREMENTS


def filter_extension_module(name, lib_objs, lib_headers):
    return Extension(
        f"thumbor.ext.filters.{name}",
        [f"thumbor/ext/filters/{name}.c"] + lib_objs,
        libraries=["m"],
        include_dirs=["thumbor/ext/filters/lib"],
        depends=["setup.py"] + lib_objs + lib_headers,
        extra_compile_args=[
            "-Wall",
            "-Wextra",
            "-Werror",
            "-Wno-unused-parameter",
        ],
    )


def gather_filter_extensions():
    files = glob.glob("thumbor/ext/filters/_*.c")
    lib_objs = glob.glob("thumbor/ext/filters/lib/*.c")
    lib_headers = glob.glob("thumbor/ext/filters/lib/*.h")

    return [
        filter_extension_module(f[0:-2].split("/")[-1], lib_objs, lib_headers)
        for f in files
    ]


def run_setup(extension_modules=None):
    if extension_modules is None:
        extension_modules = []

    if "CFLAGS" not in os.environ:
        os.environ["CFLAGS"] = ""

    setup(
        name="thumbor",
        version=__version__,
        description="thumbor is an open-source photo thumbnail service by globo.com",
        long_description=long_description,
        long_description_content_type="text/markdown",
        keywords="imaging face detection feature thumbnail imagemagick pil opencv",
        author="globo.com",
        author_email="thumbor@googlegroups.com",
        url="https://github.com/thumbor/thumbor/wiki",
        license="MIT",
        python_requires=">=3.6",
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Operating System :: MacOS",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3 :: Only",
            "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
            "Topic :: Multimedia :: Graphics :: Presentation",
        ],
        packages=["thumbor"],
        package_dir={"thumbor": "thumbor"},
        include_package_data=True,
        package_data={"": ["*.xml"]},
        install_requires=[
            "colorama==0.*,>=0.4.3",
            "derpconf==0.*,>=0.8.3",
            "libthumbor==2.*,>=2.0.2",
            "piexif==1.*,>=1.1.3",
            "Pillow>=9.0.0",
            "pytz>=2019.3.0",
            "statsd==3.*,>=3.3.0",
            "tornado==6.*,>=6.0.3",
            "thumbor-plugins-gifv==0.*,>=0.1.2",
            "webcolors==1.*,>=1.10.0",
        ],
        extras_require={
            "all": ALL_REQUIREMENTS,
            "opencv": OPENCV_REQUIREMENTS,
            "tests": ALL_REQUIREMENTS + TESTS_REQUIREMENTS,
        },
        entry_points={
            "console_scripts": [
                "thumbor=thumbor.server:main",
                "thumbor-url=thumbor.url_composer:main",
                "thumbor-config=thumbor.config:generate_config",
                "thumbor-doctor=thumbor.doctor:main",
            ],
        },
        ext_modules=extension_modules,
    )


try:
    run_setup(gather_filter_extensions())
except SystemExit as exit_error:
    print(f"\n\n{'*' * 66}")
    logging.exception(exit_error)
    print(f"\n\n{'*' * 66}")
    print(
        "Couldn't build one or more native extensions"
        ", skipping compilation.\n\n"
    )
    run_setup()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect
from tornado.testing import gen_test

import thumbor.filters
from tests.base import FilterTestCase
from thumbor.config import Config
from thumbor.context import Context, RequestParameters
from thumbor.filters.rotate import Filter
from thumbor.importer import Importer


class RotateFilterTestCase(FilterTestCase):
    @gen_test
    async def test_rotate_filter(self):
        image = await self.get_filtered(
            "source.jpg", "thumbor.filters.rotate", "rotate(180)"
        )
        expected = self.get_fixture("rotate.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.97)

    @gen_test
    async def test_rotate_filter_with_invalid_value(self):
        image = await self.get_filtered(
            "source.jpg", "thumbor.filters.rotate", "rotate(181)"
        )
        expected = self.get_fixture("source.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_equal(1)


class FakeRotateEngine:
    def __init__(self):
        self.rotate_val = None

    def rotate(self, rotate_val):
        self.rotate_val = rotate_val

    def is_multiple(self):
        return False


class FakeRotateEngineRotateFilterTestCase(FilterTestCase):
    def setUp(self):
        super().setUp()
        conf = Config()
        imp = Importer(conf)
        imp.filters = [Filter]
        self.context = Context(None, conf, imp)
        self.context.request = RequestParameters()

    @gen_test
    async def test_disallows_invalid_context_should_be_none(self):
        runner = self.context.filters_factory.create_instances(
            self.context, "rotate(91)"
        )
        filter_instances = runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        filter_instances[0].engine = FakeRotateEngine()
        await filter_instances[0].run()
        expect(filter_instances[0].engine.rotate_val).to_be_null()

    @gen_test
    async def test_should_equal_90(self):
        runner = self.context.filters_factory.create_instances(
            self.context, "rotate(90)"
        )
        filter_instances = runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        filter_instances[0].engine = FakeRotateEngine()
        await filter_instances[0].run()
        expect(filter_instances[0].engine.rotate_val).to_equal(90)

    @gen_test
    async def test_normalized_rotate_should_equal_180(self):
        runner = self.context.filters_factory.create_instances(
            self.context, "rotate(540)"
        )
        filter_instances = runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        filter_instances[0].engine = FakeRotateEngine()
        await filter_instances[0].run()
        expect(filter_instances[0].engine.rotate_val).to_equal(180)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.handlers import ContextHandler
from thumbor.utils import logger


class BlacklistHandler(ContextHandler):
    async def get(self):
        blacklist = await self.get_blacklist_contents()

        self.write(blacklist)
        self.set_header("Content-Type", "text/plain")
        self.set_status(200)

    async def put(self):
        blacklist = await self.get_blacklist_contents()
        blacklist += self.request.query + "\n"
        logger.debug("Adding to blacklist: %s", self.request.query)
        await self.context.modules.storage.put(
            "blacklist.txt", blacklist.encode()
        )
        self.set_status(200)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

BUILTIN_EXTENSIONS = [
    "thumbor.ext.filters._alpha",
    "thumbor.ext.filters._bounding_box",
    "thumbor.ext.filters._brightness",
    "thumbor.ext.filters._colorize",
    "thumbor.ext.filters._composite",
    "thumbor.ext.filters._contrast",
    "thumbor.ext.filters._convolution",
    "thumbor.ext.filters._curve",
    "thumbor.ext.filters._equalize",
    "thumbor.ext.filters._fill",
    "thumbor.ext.filters._nine_patch",
    "thumbor.ext.filters._noise",
    "thumbor.ext.filters._rgb",
    "thumbor.ext.filters._round_corner",
    "thumbor.ext.filters._saturation",
    "thumbor.ext.filters._sharpen",
]



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from unittest import TestCase, mock

from thumbor.config import Config
from thumbor.context import Context, RequestParameters
from thumbor.optimizers.jpegtran import Optimizer


class JpegtranOptimizerTest(TestCase):
    def setUp(self):
        self.patcher = mock.patch("thumbor.optimizers.jpegtran.Popen")
        self.mock_popen = self.patcher.start()
        self.os_path_exists_patcher = mock.patch(
            "thumbor.optimizers.jpegtran.exists"
        )
        self.mock_os_path_exists = self.os_path_exists_patcher.start()

    def tearDown(self):
        self.patcher.stop()
        self.os_path_exists_patcher.stop()

    @staticmethod
    def get_optimizer(filters=None, progressive=False, scans_file=""):
        conf = Config()
        conf.STATSD_HOST = ""
        conf.JPEGTRAN_PATH = "/somewhere/jpegtran"
        conf.PROGRESSIVE_JPEG = progressive
        conf.JPEGTRAN_SCANS_FILE = scans_file
        req = RequestParameters(filters=filters)
        ctx = Context(config=conf)
        ctx.request = req
        optimizer = Optimizer(ctx)

        return optimizer

    def test_should_run_for_jpeg(self):
        optimizer = self.get_optimizer()
        self.assertTrue(optimizer.should_run(".jpg", ""))
        self.assertTrue(optimizer.should_run(".jpeg", ""))

    def test_should_not_run_for_not_jpeg(self):
        optimizer = self.get_optimizer()

        self.assertFalse(optimizer.should_run(".png", ""))
        self.assertFalse(optimizer.should_run(".webp", ""))
        self.assertFalse(optimizer.should_run(".gif", ""))

    def test_should_not_run_if_binary_jpegtran_path_does_not_exist(self):
        self.mock_os_path_exists.return_value = False

        optimizer = self.get_optimizer()
        self.assertFalse(optimizer.should_run(".jpg", ""))

    def test_should_optimize(self):
        input_buffer = "input buffer"
        output_buffer = "output buffer"
        self.mock_popen.return_value.returncode = 0
        self.mock_popen.return_value.communicate.return_value = (
            output_buffer,
            "Error",
        )

        optimizer = self.get_optimizer()
        return_buffer = optimizer.run_optimizer(".jpg", input_buffer)

        self.mock_popen.return_value.communicate.assert_called_with(
            input_buffer
        )
        self.assertEqual(output_buffer, return_buffer)

    def test_should_return_old_buffer_for_invalid_extension(self):
        optimizer = self.get_optimizer()
        buffer = "garbage"

        return_buffer = optimizer.run_optimizer(".png", buffer)

        self.assertEqual(return_buffer, buffer)

    def test_should_return_old_buffer_for_invalid_image(self):
        optimizer = self.get_optimizer()
        buffer = "garbage"

        self.mock_popen.return_value.returncode = 1
        self.mock_popen.return_value.communicate.return_value = (
            "Output",
            "Error",
        )

        return_buffer = optimizer.run_optimizer(".jpg", buffer)

        self.assertEqual(return_buffer, buffer)

    def test_should_preserve_comments_if_strip_icc_filter_set(self):
        self.mock_popen.return_value.returncode = 0
        self.mock_popen.return_value.communicate.return_value = (
            "Output",
            "Error",
        )

        optimizer = self.get_optimizer(filters=["strip_icc"])
        optimizer.run_optimizer(".jpg", "")

        command = self.mock_popen.call_args[0][0]

        self.assertIn("-copy", command)
        self.assertIn("comments", command)
        self.assertNotIn("all", command)

        optimizer = self.get_optimizer()
        optimizer.run_optimizer(".jpg", "")

        command = self.mock_popen.call_args[0][0]

        self.assertIn("-copy", command)
        self.assertIn("all", command)
        self.assertNotIn("comments", command)

    def test_should_make_progressive_when_configured(self):
        self.mock_popen.return_value.returncode = 0
        self.mock_popen.return_value.communicate.return_value = (
            "Output",
            "Error",
        )

        optimizer = self.get_optimizer(progressive=False)
        optimizer.run_optimizer(".jpg", "")

        args, _ = self.mock_popen.call_args
        command = args[0]

        self.assertNotIn("-progressive", command)

        optimizer = self.get_optimizer(progressive=True)
        optimizer.run_optimizer(".jpg", "")

        args, _ = self.mock_popen.call_args
        command = args[0]

        self.assertIn("-progressive", command)

    def test_should_not_use_scans_file_when_not_configured(self):
        self.mock_popen.return_value.returncode = 0
        self.mock_popen.return_value.communicate.return_value = (
            "Output",
            "Error",
        )

        optimizer = self.get_optimizer(scans_file="")
        optimizer.run_optimizer(".jpg", "")

        args, _ = self.mock_popen.call_args
        command = args[0]

        self.assertNotIn("-scans", command)

    def test_should_use_scans_file_when_configured_and_exists(self):
        self.mock_popen.return_value.returncode = 0
        self.mock_popen.return_value.communicate.return_value = (
            "Output",
            "Error",
        )

        self.mock_os_path_exists.return_value = True

        optimizer = self.get_optimizer(scans_file="scans_test.txt")
        optimizer.run_optimizer(".jpg", "")

        args, _ = self.mock_popen.call_args
        command = args[0]

        self.assertIn("-scans", command)
        self.assertIn("scans_test.txt", command)

    @mock.patch("thumbor.optimizers.jpegtran.logger.warning")
    def test_should_log_warning_when_scans_file_missing(self, warn_logger):
        self.mock_popen.return_value.returncode = 0
        self.mock_popen.return_value.communicate.return_value = (
            "Output",
            "Error",
        )

        self.mock_os_path_exists.side_effect = (
            lambda filename: filename != "scans_test.txt"
        )

        optimizer = self.get_optimizer(scans_file="scans_test.txt")
        optimizer.run_optimizer(".jpg", "")

        args, _ = self.mock_popen.call_args
        command = args[0]

        self.assertNotIn("-scans", command)
        warn_logger.assert_called_once()

    @mock.patch("thumbor.optimizers.jpegtran.logger.warning")
    def test_should_log_warning_when_failed(self, warn_logger):
        optimizer = self.get_optimizer()

        self.mock_popen.return_value.returncode = 1
        self.mock_popen.return_value.communicate.return_value = (
            "Output",
            "Error",
        )

        optimizer.run_optimizer(".jpg", "garbage")

        warn_logger.assert_called_once()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com timehome@corp.globo.com

# This file is copied from legacy way of using thumbor result storages,
# thus we did not fix lint errors

# pylint: disable=all

from datetime import datetime
from os.path import abspath, dirname, exists, getmtime, join
from shutil import move
from uuid import uuid4

import pytz

from thumbor.engines import BaseEngine
from thumbor.result_storages import BaseStorage, ResultStorageResult
from thumbor.utils import deprecated, logger


class Storage(BaseStorage):
    PATH_FORMAT_VERSION = "v2"

    @property
    def is_auto_webp(self):
        return (
            self.context.config.AUTO_WEBP and self.context.request.accepts_webp
        )

    def put(self, bytes):
        file_abspath = self.normalize_path(self.context.request.url)
        if not self.validate_path(file_abspath):
            logger.warn(
                "[RESULT_STORAGE] unable to write outside root path: %s"
                % file_abspath
            )
            return
        temp_abspath = "%s.%s" % (file_abspath, str(uuid4()).replace("-", ""))
        file_dir_abspath = dirname(file_abspath)
        logger.debug(
            "[RESULT_STORAGE] putting at %s (%s)"
            % (file_abspath, file_dir_abspath)
        )

        self.ensure_dir(file_dir_abspath)

        with open(temp_abspath, "wb") as _file:
            _file.write(bytes)

        move(temp_abspath, file_abspath)

    def get(self, callback):
        path = self.context.request.url
        file_abspath = self.normalize_path(path)
        if not self.validate_path(file_abspath):
            logger.warn(
                "[RESULT_STORAGE] unable to read from outside root path: %s"
                % file_abspath
            )
            return None
        logger.debug("[RESULT_STORAGE] getting from %s" % file_abspath)

        if not exists(file_abspath) or self.is_expired(file_abspath):
            logger.debug(
                "[RESULT_STORAGE] image not found at %s" % file_abspath
            )
            callback(None)
        else:
            with open(file_abspath, "rb") as f:
                buffer = f.read()

            result = ResultStorageResult(
                buffer=buffer,
                metadata={
                    "LastModified": datetime.fromtimestamp(
                        getmtime(file_abspath)
                    ).replace(tzinfo=pytz.utc),
                    "ContentLength": len(buffer),
                    "ContentType": BaseEngine.get_mimetype(buffer),
                },
            )

            callback(result)

    def validate_path(self, path):
        return abspath(path).startswith(
            self.context.config.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH
        )

    def normalize_path(self, path):
        path_segments = [
            self.context.config.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH.rstrip(
                "/"
            ),
            Storage.PATH_FORMAT_VERSION,
        ]
        if self.is_auto_webp:
            path_segments.append("webp")

        path_segments.extend(
            [
                self.partition(path),
                path.lstrip("/"),
            ]
        )

        normalized_path = join(*path_segments).replace("http://", "")
        return normalized_path

    def partition(self, path_raw):
        path = path_raw.lstrip("/")
        return join("".join(path[0:2]), "".join(path[2:4]))

    def is_expired(self, path):
        expire_in_seconds = self.context.config.get(
            "RESULT_STORAGE_EXPIRATION_SECONDS", None
        )

        if expire_in_seconds is None or expire_in_seconds == 0:
            return False

        timediff = datetime.now() - datetime.fromtimestamp(getmtime(path))
        return timediff.seconds > expire_in_seconds

    @deprecated("Use result's last_modified instead")
    def last_updated(self):
        path = self.context.request.url
        file_abspath = self.normalize_path(path)
        if not self.validate_path(file_abspath):
            logger.warn(
                "[RESULT_STORAGE] unable to read from outside root path: %s"
                % file_abspath
            )
            return True
        logger.debug("[RESULT_STORAGE] getting from %s" % file_abspath)

        if not exists(file_abspath) or self.is_expired(file_abspath):
            logger.debug(
                "[RESULT_STORAGE] image not found at %s" % file_abspath
            )
            return True

        return datetime.fromtimestamp(getmtime(file_abspath)).replace(
            tzinfo=pytz.utc
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect

import thumbor.handler_lists.healthcheck as healthcheck_handler_list
from tests.base import TestCase
from thumbor.config import Config
from thumbor.detectors import feature_detector
from thumbor.detectors.face_detector import Detector as face_detector
from thumbor.engines.gif import Engine as gif_engine
from thumbor.engines.pil import Engine as pil_engine
from thumbor.filters.rgb import Filter as rgb_filter
from thumbor.importer import Importer
from thumbor.loaders import http_loader
from thumbor.result_storages.file_storage import Storage as result_file_storage
from thumbor.storages.file_storage import Storage as file_storage


class ImporterTestCase(TestCase):
    def get_config(self):
        return Config(
            ENGINE="thumbor.engines.pil",
            GIF_ENGINE="thumbor.engines.gif",
            LOADER="thumbor.loaders.http_loader",
            STORAGE="thumbor.storages.file_storage",
            UPLOAD_PHOTO_STORAGE="thumbor.storages.file_storage",
            RESULT_STORAGE="thumbor.result_storages.file_storage",
            DETECTORS=["thumbor.detectors.face_detector"],
            FILTERS=["thumbor.filters.rgb"],
            HANDLER_LISTS=["thumbor.handler_lists.healthcheck"],
        )

    def test_import_item_should_be_proper_item(self):
        importer = Importer(self.get_config())
        importer.import_modules()
        data = {
            "ENGINE": pil_engine,
            "GIF_ENGINE": gif_engine,
            "LOADER": http_loader,
            "STORAGE": file_storage,
            "UPLOAD_PHOTO_STORAGE": file_storage,
            "RESULT_STORAGE": result_file_storage,
            "DETECTORS": (face_detector,),
            "FILTERS": (rgb_filter,),
            "HANDLER_LISTS": (healthcheck_handler_list,),
        }

        for key, value in data.items():
            prop, default_value = (None, None)
            if hasattr(importer, key.lower()):
                prop, default_value = (getattr(importer, key.lower()), value)

            if prop is tuple:
                for index, item in enumerate(prop):
                    expect(item).not_to_be_null().to_equal(
                        default_value[index]
                    )
            else:
                expect(prop).not_to_be_null().to_equal(default_value)

    @staticmethod
    def test_single_item_should_equal_file_storage():
        importer = Importer(None)
        importer.import_item(
            config_key="file_storage",
            item_value="thumbor.storages.file_storage",
            class_name="Storage",
        )
        expect(getattr(importer, "file_storage")).to_equal(file_storage)

    @staticmethod
    def test_multiple_items_can_be_imported():
        importer = Importer(None)
        importer.import_item(
            config_key="detectors",
            is_multiple=True,
            item_value=(
                "thumbor.detectors.feature_detector",
                "thumbor.detectors.feature_detector",
            ),
        )

        expect(importer.detectors).to_length(2)
        expect(importer.detectors).to_include(feature_detector)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com timehome@corp.globo.com

import importlib
import sys

# For backward compatibility
from libthumbor.url_signers import BaseUrlSigner  # NOQA

sys.modules[__name__] = importlib.import_module(
    "libthumbor.url_signers"
)  # NOQA



from os.path import join

from tornado.testing import gen_test

from . import EngineCase
from .urls_helpers import MAX_DATASET_SIZE, UrlsTester, single_dataset


class PILTest(EngineCase):
    engine = "thumbor.engines.pil"

    @gen_test(timeout=MAX_DATASET_SIZE * 5)
    async def test_single_params(self):
        if not self._app:
            return True
        group = list(single_dataset())
        count = len(group)
        tester = UrlsTester(self.http_client)

        print("Requests count: %d" % count)
        for options in group:
            joined_parts = join(*options)
            url = "unsafe/%s" % joined_parts
            await tester.try_url(self.get_url(f"/{url}"))

        tester.report()

    # def test_combined_params__with_pil(self):
    #     if not self._app:
    #         return True
    #     combined_dataset(self.retrieve)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from os.path import abspath, dirname, join

from preggy import expect
from tornado.testing import gen_test

from tests.base import TestCase
from thumbor.config import Config
from thumbor.context import Context
from thumbor.importer import Importer
from thumbor.loaders import LoaderResult

STORAGE_PATH = abspath(join(dirname(__file__), "../fixtures/images/"))


class CompatibilityLoaderTestCase(TestCase):
    def get_context(self):
        config = Config(
            FILE_LOADER_ROOT_PATH=STORAGE_PATH,
            COMPATIBILITY_LEGACY_LOADER="tests.compatibility.legacy_file_loader",
        )
        importer = Importer(config)
        importer.import_modules()
        return Context(config=config, importer=importer)

    async def load_file(self, context, file_name):
        Importer.deprecated_monkey_patch_tornado_return_future()
        from thumbor.compatibility.loader import (  # pylint: disable=import-outside-toplevel
            load,
        )

        return await load(context, file_name)

    @gen_test
    async def test_should_raise_for_invalid_compatibility_loader(self):
        config = Config(
            FILE_LOADER_ROOT_PATH=STORAGE_PATH,
        )
        importer = Importer(config)
        importer.import_modules()
        ctx = Context(config=config, importer=importer)

        with expect.error_to_happen(
            RuntimeError,
            message=(
                "The 'COMPATIBILITY_LEGACY_LOADER' configuration should point "
                "to a valid loader when using compatibility loader."
            ),
        ):
            await self.load_file(ctx, "image.jpg")

    @gen_test
    async def test_should_load_file(self):
        result = await self.load_file(self.context, "image.jpg")
        expect(result).to_be_instance_of(LoaderResult)
        with open(join(STORAGE_PATH, "image.jpg"), "rb") as img:
            expect(result.buffer).to_equal(img.read())
            expect(result.successful).to_be_true()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from shutil import which

from preggy import expect
from tornado.testing import gen_test

from tests.handlers.test_base_handler import BaseImagingTestCase
from thumbor.config import Config
from thumbor.context import Context, ServerParameters
from thumbor.importer import Importer

# pylint: disable=broad-except,abstract-method,attribute-defined-outside-init,line-too-long,too-many-public-methods
# pylint: disable=too-many-lines


class ImageOperationsImageCoverTestCase(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.AUTO_WEBP = True
        cfg.USE_GIFSICLE_ENGINE = True

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        ctx = Context(server, cfg, importer)
        ctx.server.gifsicle_path = which("gifsicle")
        return ctx

    @gen_test
    async def test_can_get_image_cover(self):
        response = await self.async_fetch(
            "/unsafe/filters:cover()/animated.gif"
        )

        expect(response.code).to_equal(200)
        expect(response.headers["Content-Type"]).to_equal("image/gif")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

# pylint: disable=no-member
from os.path import abspath, dirname, join

from thumbor.filters import BaseFilter, filter_method
from thumbor.utils import logger

try:
    import cv2
    import numpy as np

    OPENCV_AVAILABLE = True
except ImportError:
    OPENCV_AVAILABLE = False


CASCADE_FILE_PATH = abspath(join(dirname(__file__), "haarcascade_eye.xml"))

MIN_SIZE = (20, 20)
HAAR_SCALE = 1.2
MIN_NEIGHBORS = 3
RED_THRESHOLD = 2.0


class Filter(BaseFilter):
    @filter_method()
    async def red_eye(self) -> None:
        if not OPENCV_AVAILABLE:
            logger.error(
                "Can't use red eye removal filter if OpenCV and NumPy are not available."
            )

            return

        faces = [
            face
            for face in self.context.request.focal_points
            if face.origin == "Face Detection"
        ]

        if not faces:
            return

        mode, data = self.engine.image_data_as_rgb()
        mode = mode.lower()
        size = self.engine.size

        image = np.ndarray(
            shape=(size[1], size[0], 4 if mode == "rgba" else 3),
            dtype="|u1",
            buffer=data,
        ).copy()

        for face in faces:
            face_x = int(face.x - face.width / 2)
            face_y = int(face.y - face.height / 2)

            face_image = image[
                face_y : face_y + face.height, face_x : face_x + face.width
            ]

            eye_rects = self.cascade.detectMultiScale(
                face_image,
                scaleFactor=HAAR_SCALE,
                minNeighbors=MIN_NEIGHBORS,
                minSize=MIN_SIZE,
            )

            for pos_x, pos_y, width, height in eye_rects:
                # Crop the eye region
                eye_image = face_image[
                    pos_y : pos_y + height, pos_x : pos_x + width
                ]

                # split the images into 3 channels
                red, green, blue = cv2.split(eye_image)

                # Add blue and green channels
                blue_green = cv2.add(blue, green)
                mean = blue_green // 2
                # threshold the mask based on red color and combination of blue and green color
                mask = ((red > RED_THRESHOLD * mean) & (red > 60)).astype(
                    np.uint8
                ) * 255
                # Some extra region may also get detected , we find the largest region
                # find all contours
                contours_return = cv2.findContours(
                    mask.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE
                )  # It return contours and Hierarchy

                if len(contours_return) == 2:
                    contours, _ = contours_return
                else:
                    _, contours, _ = contours_return

                # find contour with max area
                max_area = 0
                max_cont = None

                for cont in contours:
                    area = cv2.contourArea(cont)

                    if area > max_area:
                        max_area = area
                        max_cont = cont

                if max_cont is None:
                    continue

                mask = mask * 0  # Reset the mask image to complete black image
                # draw the biggest contour on mask
                cv2.drawContours(mask, [max_cont], 0, (255), -1)
                # Close the holes to make a smooth region
                mask = cv2.morphologyEx(
                    mask,
                    cv2.MORPH_CLOSE,
                    cv2.getStructuringElement(cv2.MORPH_DILATE, (5, 5)),
                )
                mask = cv2.dilate(mask, (3, 3), iterations=3)

                # The information of only red color is lost,
                # So we fill the mean of blue and green color in all
                # three channels(BGR) to maintain the texture
                # Fill this black mean value to masked image
                mean = cv2.bitwise_and(mean, mask)  # mask the mean image
                mean = cv2.cvtColor(
                    mean, cv2.COLOR_GRAY2RGB
                )  # convert mean to 3 channel
                mask = cv2.cvtColor(
                    mask, cv2.COLOR_GRAY2RGB
                )  # convert mask to 3 channel
                eye = (
                    cv2.bitwise_and(~mask, eye_image) + mean
                )  # Copy the mean color to masked region to color image
                face_image[pos_y : pos_y + height, pos_x : pos_x + width] = eye

        self.engine.set_image_data(image.tobytes())

    @property
    def cascade(self) -> None:
        if not hasattr(self, "_cascade"):
            setattr(self, "_cascade", cv2.CascadeClassifier(CASCADE_FILE_PATH))

        return getattr(self, "_cascade")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect
from tornado.testing import gen_test

from tests.base import FilterTestCase


class RoundCornerFilterTestCase(FilterTestCase):
    @gen_test
    async def test_round_corner_filter_with_a_radius(self):
        image = await self.get_filtered(
            "source.jpg",
            "thumbor.filters.round_corner",
            "round_corner(50,255,0,0)",
        )
        expected = self.get_fixture("round_corner.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_round_corner_filter_with_ab_radius(self):
        image = await self.get_filtered(
            "source.jpg",
            "thumbor.filters.round_corner",
            "round_corner(50|75,255,0,0)",
        )
        expected = self.get_fixture("round_corner1.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_round_corner_filter_with_blue_background(self):
        image = await self.get_filtered(
            "source.jpg",
            "thumbor.filters.round_corner",
            "round_corner(50,0,0,255)",
        )
        expected = self.get_fixture("round_corner2.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_round_corner_filter_without_transparent(self):
        image = await self.get_filtered(
            "source.jpg",
            "thumbor.filters.round_corner",
            "round_corner(50,255,0,0,0)",
        )
        expected = self.get_fixture("round_corner3.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_round_corner_filter_with_transparent(self):
        image = await self.get_filtered(
            "source.jpg",
            "thumbor.filters.round_corner",
            "round_corner(50,255,0,0,1)",
        )
        expected = self.get_fixture("round_corner3.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_round_corner_filter_with_transparent_by_png(self):
        image = await self.get_filtered(
            "source.jpg",
            "thumbor.filters.round_corner",
            "round_corner(50,255,0,0,1)",
            mode="RGBA",
        )
        expected = self.get_fixture("round_corner.png", mode="RGBA")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_round_corner_filter_with_transparent_by_webp(self):
        image = await self.get_filtered(
            "source.jpg",
            "thumbor.filters.round_corner",
            "round_corner(50,255,0,0,1)",
            mode="RGBA",
        )
        expected = self.get_fixture("round_corner.webp", mode="RGBA")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import tempfile
from os.path import abspath, dirname, join

from preggy import expect
from tornado.testing import gen_test

from tests.base import TestCase
from tests.fixtures.transformer_test_data import (
    FIT_IN_CROP_DATA,
    TESTITEMS,
    MockErrorSyncDetector,
    MockSyncDetector,
    TestData,
)
from thumbor.config import Config
from thumbor.context import Context, ServerParameters
from thumbor.importer import Importer
from thumbor.transformer import Transformer


class TransformerTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.root_folder = (
            tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
        )
        cls.root_path = cls.root_folder.name
        cls.loader_path = abspath(
            join(dirname(__file__), "../fixtures/images/")
        )
        cls.base_uri = "/image"

    @classmethod
    def tearDownClass(cls):
        if cls.root_folder is not None:
            cls.root_folder.cleanup()

    def setUp(self):
        super().setUp()
        self.has_handled = False

    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.STORAGE = "thumbor.storages.file_storage"
        cfg.FILE_STORAGE_ROOT_PATH = self.root_path

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        return Context(server, cfg, importer)

    @gen_test
    async def test_invalid_crop(self):
        data = TestData(
            source_width=800,
            source_height=600,
            target_width=800,
            target_height=600,
            halign="right",
            valign="top",
            focal_points=[],
            crop_left=200,
            crop_top=0,
            crop_right=100,
            crop_bottom=100,
        )

        ctx = data.to_context()
        engine = ctx.modules.engine

        trans = Transformer(ctx)

        await trans.transform()
        expect(engine.calls["crop"]).to_be_empty()

    @staticmethod
    def validate_resize(test_data):
        expect(test_data).to_be_resized()
        expect(test_data).to_be_cropped()

    @gen_test
    async def test_can_resize_images(self):
        for item in TESTITEMS:
            context = item.to_context()
            trans = Transformer(context)
            await trans.transform()
            self.validate_resize(item)

    @gen_test
    async def test_can_resize_images_with_detectors(self):
        for item in TESTITEMS:
            context = item.to_context(detectors=[MockSyncDetector])
            trans = Transformer(context)
            await trans.transform()
            self.validate_resize(item)

    @gen_test
    async def test_can_resize_images_with_detection_error(self):
        test_data = TestData(
            source_width=800,
            source_height=600,
            target_width=400,
            target_height=150,
            halign="center",
            valign="middle",
            focal_points=[],
            crop_left=0,
            crop_top=75,
            crop_right=800,
            crop_bottom=375,
        )
        context = test_data.to_context(
            detectors=[MockErrorSyncDetector], ignore_detector_error=True
        )
        trans = Transformer(context)
        await trans.transform()
        self.validate_resize(test_data)

    @gen_test
    async def test_can_resize_images_with_detection_error_not_ignoring_it(
        self,
    ):
        test_data = TestData(
            source_width=800,
            source_height=600,
            target_width=400,
            target_height=150,
            halign="center",
            valign="middle",
            focal_points=[],
            crop_left=0,
            crop_top=75,
            crop_right=800,
            crop_bottom=375,
        )
        context = test_data.to_context(
            detectors=[MockErrorSyncDetector], ignore_detector_error=False
        )
        trans = Transformer(context)

        with expect.error_to_happen(Exception, message="x"):
            await trans.transform()

        expect(test_data.engine.calls["resize"]).to_length(0)

    @gen_test
    async def test_can_fit_in(self):
        for (test_data, (width, height, should_resize)) in FIT_IN_CROP_DATA:
            context = test_data.to_context()
            engine = context.modules.engine
            trans = Transformer(context)

            await trans.transform()

            expect(engine.calls["crop"]).to_be_empty()
            if should_resize:
                expect(engine.calls["resize"]).to_length(1)
                expect(engine.calls["resize"][0]["width"]).to_equal(width)
                expect(engine.calls["resize"][0]["height"]).to_equal(height)
            else:
                expect(engine.calls["resize"]).to_be_empty()

    @gen_test
    async def test_can_transform_meta_with_orientation(self):
        data = TestData(
            source_width=800,
            source_height=600,
            target_width=100,
            target_height=100,
            halign="right",
            valign="top",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
            meta=True,
        )

        ctx = data.to_context()
        ctx.config.RESPECT_ORIENTATION = True
        engine = ctx.modules.engine

        trans = Transformer(ctx)
        await trans.transform()
        expect(engine.calls["reorientate"]).to_equal(1)

    @gen_test
    async def test_can_transform_with_flip(self):
        data = TestData(
            source_width=800,
            source_height=600,
            target_width=-800,
            target_height=-600,
            halign="right",
            valign="top",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
        )
        ctx = data.to_context()
        engine = ctx.modules.engine
        trans = Transformer(ctx)

        await trans.transform()

        expect(engine.calls["horizontal_flip"]).to_equal(1)
        expect(engine.calls["vertical_flip"]).to_equal(1)

    @gen_test
    async def test_can_resize_with_stretch(self):
        data = TestData(
            source_width=800,
            source_height=600,
            target_width=800,
            target_height=200,
            halign="right",
            valign="top",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
            stretch=True,
        )
        ctx = data.to_context()
        engine = ctx.modules.engine
        trans = Transformer(ctx)

        await trans.transform()

        expect(engine.calls["resize"]).to_equal(
            [{"width": 800, "height": 200}]
        )
        expect(engine.calls["crop"]).to_be_empty()

    @gen_test
    async def test_can_extract_cover(self):
        data = TestData(
            source_width=800,
            source_height=600,
            target_width=-800,
            target_height=-600,
            halign="right",
            valign="top",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
        )
        ctx = data.to_context()
        ctx.request.filters = "cover()"
        ctx.request.image = "some.gif"
        ctx.request.extension = "GIF"
        ctx.request.engine.extension = ".gif"
        ctx.config.USE_GIFSICLE_ENGINE = True
        engine = ctx.modules.engine
        trans = Transformer(ctx)

        await trans.transform()

        expect(engine.calls["cover"]).to_equal(1)

    @gen_test
    async def test_get_target_dimensions(self):
        data = TestData(
            source_width=800,
            source_height=600,
            target_width=600,
            target_height=400,
            halign="right",
            valign="top",
            focal_points=[],
            crop_left=200,
            crop_top=0,
            crop_right=100,
            crop_bottom=100,
        )

        ctx = data.to_context()
        trans = Transformer(ctx)
        dimensions = trans.get_target_dimensions()
        expect(dimensions).to_equal((600, 400))

        await trans.transform()

        expect(dimensions).to_equal((600, 400))



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

try:
    import numpy as np
    import cv2
except ImportError:
    pass

from thumbor.detectors import BaseDetector
from thumbor.point import FocalPoint
from thumbor.utils import logger


class Detector(BaseDetector):
    async def detect(self):
        if not self.verify_cv():
            await self.next()

            return

        engine = self.context.modules.engine
        try:
            img = np.array(
                engine.convert_to_grayscale(update_image=False, alpha=False)
            )
        except Exception as error:
            logger.exception(error)
            logger.warning(
                "Error during feature detection; skipping to next detector"
            )

            return await self.next()  # pylint: disable=not-callable

        points = cv2.goodFeaturesToTrack(  # pylint: disable=no-member
            img,
            maxCorners=20,
            qualityLevel=0.04,
            minDistance=1.0,
            useHarrisDetector=False,
        )

        if points is not None:
            for point in points:
                x_pos, y_pos = point.ravel()
                self.context.request.focal_points.append(
                    FocalPoint(x_pos.item(), y_pos.item(), 1)
                )

            return

        await self.next()  # pylint: disable=not-callable



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import logging
from functools import wraps
from io import BytesIO
from PIL import Image

try:
    from PIL import ImageCms  # pylint: disable=ungrouped-imports
except ImportError:
    ImageCms = None
    DEFAULT_SRGB_PROFILE = None
    TRANSFORM_FLAGS = 0
else:
    DEFAULT_SRGB_PROFILE = ImageCms.ImageCmsProfile(
        ImageCms.createProfile("sRGB")
    )
    TRANSFORM_FLAGS = (
        ImageCms.FLAGS["NOTCACHE"]
        | ImageCms.FLAGS["NOTPRECALC"]
        | ImageCms.FLAGS["BLACKPOINTCOMPENSATION"]
        | ImageCms.FLAGS["HIGHRESPRECALC"]
    )


CONTENT_TYPE = {
    ".jpg": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".gif": "image/gif",
    ".png": "image/png",
    ".webp": "image/webp",
    ".mp4": "video/mp4",
    ".webm": "video/webm",
    ".svg": "image/svg+xml",
    ".tif": "image/tiff",
    ".tiff": "image/tiff",
    ".avif": "image/avif",
    ".heic": "image/heif",
    ".heif": "image/heif",
}

EXTENSION = {
    "image/jpeg": ".jpg",
    "image/gif": ".gif",
    "image/png": ".png",
    "image/webp": ".webp",
    "video/mp4": ".mp4",
    "video/webm": ".webm",
    "image/svg+xml": ".svg",
    "image/tiff": ".tif",
    "image/avif": ".avif",
    "image/heif": ".heic",
}


logger = logging.getLogger("thumbor")  # pylint: disable=invalid-name


def deprecated(message):
    def decorator_deprecated(func):
        @wraps(func)
        def wrapper_deprecated(*args, **kwargs):
            logger.warning(
                "Deprecated function %s%s",
                func.__name__,
                message,
            )
            return func(*args, **kwargs)

        return wrapper_deprecated

    return decorator_deprecated


def get_color_space(img):
    icc = img.info.get("icc_profile")
    if not icc:
        return "RGB"

    if ImageCms is None:
        return None

    buf = BytesIO(icc)
    try:
        orig_profile = ImageCms.ImageCmsProfile(buf)
        color_space = orig_profile.profile.xcolor_space
    except (AttributeError, OSError, TypeError, ValueError):
        return None
    else:
        return color_space.strip()


def ensure_srgb(img, srgb_profile=None):
    """
    Ensures that an image either has no ICC profile (and so is implicitly
    sRGB) or has an sRGB color profile. If the image is sRGB, it is returned
    unchanged. If it has a CMYK or Gray color profile, this function will
    return an image converted to sRGB. Any color profiles in other color
    spaces will return None.
    """
    img_info = dict(img.info)
    icc = img_info.pop("icc_profile", None)
    if not icc:
        return img

    if ImageCms is None:
        raise RuntimeError("ImageCms is required for color profile utilities")

    if srgb_profile is not None:
        srgb_profile = ImageCms.ImageCmsProfile(srgb_profile)
    else:
        srgb_profile = DEFAULT_SRGB_PROFILE

    buf = BytesIO(icc)
    try:
        orig_profile = ImageCms.ImageCmsProfile(buf)
        color_space = orig_profile.profile.xcolor_space
    except (AttributeError, OSError, TypeError, ValueError):
        return None
    finally:
        buf.close()

    if color_space == "RGB ":
        logger.debug("Returning img (RGB)")
        return img

    if color_space not in ("GRAY", "CMYK"):
        # Other color spaces are rare, but best not to try to convert them.
        # Upstream understands a None return as meaning it should not
        # use it for the target encoder.
        logger.debug(
            "Cannot convert to sRGB; color space = %s",
            (color_space.strip()),
        )
        return None

    # Probably not possible to have an animated image with CMYK or GRAY icc
    # profile, but best leave it alone if we have one
    if getattr(img, "is_animated", False):
        return None

    if color_space == "GRAY":
        pil_mode = "L"
    else:
        pil_mode = "CMYK"

    logger.debug("Converting from %s to sRGB", color_space)

    transform = ImageCms.ImageCmsTransform(
        orig_profile,
        srgb_profile,
        pil_mode,
        "RGBA",
        intent=ImageCms.INTENT_RELATIVE_COLORIMETRIC,
        flags=TRANSFORM_FLAGS,
    )

    src_im = Image.new(pil_mode, img.size, "white")
    src_im.paste(img)

    dst_im = Image.new("RGBA", img.size, "white")
    dst_im.info = img_info
    dst_im = transform.apply(src_im, dst_im)
    dst_im = dst_im.convert("RGB")
    dst_im.info = img_info
    return dst_im



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from os.path import abspath, dirname, join

from preggy import expect
from tornado.testing import gen_test

from tests.base import TestCase
from thumbor.compatibility.result_storage import Storage
from thumbor.config import Config
from thumbor.context import Context, RequestParameters, ServerParameters
from thumbor.importer import Importer

STORAGE_PATH = abspath(join(dirname(__file__), "../fixtures/images/"))


class CompatibilityResultStorageTestCase(TestCase):
    def get_image_path(self, name):
        return f"./tests/fixtures/images/{name}"

    def get_image_bytes(self, name):
        with open(self.get_image_path(name), "rb") as img:
            return img.read()

    def get_image_url(self, name):
        return f"s.glbimg.com/some/{name}"

    def get_context(self):
        config = Config(
            FILE_LOADER_ROOT_PATH=STORAGE_PATH,
            COMPATIBILITY_LEGACY_RESULT_STORAGE="tests.compatibility.legacy_file_result_storage",
            STORES_CRYPTO_KEY_FOR_EACH_IMAGE=True,
        )
        self.importer = Importer(config)
        self.importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        return Context(server, config=config, importer=self.importer)

    @gen_test
    async def test_should_raise_for_invalid_compatibility_storage(self):
        request = RequestParameters(
            url="/image.jpg",
        )
        config = Config(
            FILE_LOADER_ROOT_PATH=STORAGE_PATH,
            STORES_CRYPTO_KEY_FOR_EACH_IMAGE=True,
        )
        importer = Importer(config)
        importer.import_modules()
        image_bytes = self.get_image_bytes("image.jpg")
        ctx = Context(self.context.server, config, importer)
        ctx.request = request
        storage = Storage(ctx)

        with expect.error_to_happen(
            RuntimeError,
            message=(
                "The 'COMPATIBILITY_LEGACY_RESULT_STORAGE' configuration should point "
                "to a valid result storage when using compatibility result storage."
            ),
        ):
            await storage.get()

        with expect.error_to_happen(
            RuntimeError,
            message=(
                "The 'COMPATIBILITY_LEGACY_RESULT_STORAGE' configuration should point "
                "to a valid result storage when using compatibility result storage."
            ),
        ):
            await storage.put(image_bytes)

    @gen_test
    async def test_should_put_and_get(self):
        request = RequestParameters(
            url="/image.jpg",
        )
        image_bytes = self.get_image_bytes("image.jpg")
        ctx = Context(self.context.server, self.context.config, self.importer)
        ctx.request = request
        storage = Storage(ctx)

        await storage.put(image_bytes)

        result = await storage.get()
        expect(result).not_to_be_null()
        expect(result).not_to_be_an_error()
        expect(result.successful).to_be_true()
        expect(result.buffer).to_equal(image_bytes)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.loaders import LoaderResult, file_loader, http_loader


async def load(context, path):
    # First attempt to load with file_loader
    result = await file_loader.load(context, path)

    if result.successful:
        return result

    # If file_loader failed try http_loader

    if not http_loader.validate(context, path):
        result = LoaderResult()
        result.successful = False
        result.error = LoaderResult.ERROR_BAD_REQUEST
        result.extras["reason"] = "Unallowed domain"
        result.extras["source"] = path

        return result

    return await http_loader.load(context, path)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.filters import BaseFilter, filter_method
from thumbor.utils import logger

ALLOWED_FORMATS = ["png", "jpeg", "jpg", "gif", "webp", "avif", "heic", "heif"]


class Filter(BaseFilter):
    @filter_method(BaseFilter.String)
    async def format(self, file_format):
        if file_format.lower() not in ALLOWED_FORMATS:
            logger.debug("Format not allowed: %s", file_format.lower())
            self.context.request.format = None
        else:
            logger.debug("Format specified: %s", file_format.lower())
            self.context.request.format = file_format.lower()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect

import thumbor.handler_lists.healthcheck as handler_list
from thumbor.handlers.healthcheck import HealthcheckHandler
from thumbor.testing import TestCase


class HealthcheckHandlerListTestCase(TestCase):
    def test_can_get_handlers(self):
        handlers = handler_list.get_handlers(self.context)

        expect(handlers).not_to_be_null()
        expect(handlers).to_length(1)
        url, handler = handlers[0]
        expect(url).to_equal(r"/healthcheck/?")
        expect(handler).to_equal(HealthcheckHandler)

    def test_can_get_handlers_with_custom_url(self):
        ctx = self.get_context()
        ctx.config.HEALTHCHECK_ROUTE = "/health"

        handlers = handler_list.get_handlers(ctx)

        expect(handlers).not_to_be_null()
        expect(handlers).to_length(1)
        url, handler = handlers[0]
        expect(url).to_equal("/health")
        expect(handler).to_equal(HealthcheckHandler)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import sys
from unittest.mock import patch

import pytest
from preggy import expect
from tornado.testing import gen_test

import thumbor.filters
from tests.base import TestCase
from thumbor.config import Config
from thumbor.context import Context
from thumbor.filters import BaseFilter, FiltersFactory, filter_method
from thumbor.importer import Importer

FILTER_PARAMS_DATA = [
    {
        "type": BaseFilter.Number,
        "values": [
            ("1", 1),
            ("10", 10),
            ("99", 99),
            ("-1", -1),
            ("-10", -10),
            ("010", 10),
            ("  1 ", 1),
            ("0", 0),
        ],
        "invalid_values": ["x", "x10", "10x", "- 1", ""],
    },
    {
        "type": BaseFilter.PositiveNumber,
        "values": [
            ("1", 1),
            ("10", 10),
            ("99", 99),
            (" 1 ", 1),
            ("010", 10),
            ("0", 0),
        ],
        "invalid_values": ["-1", "x", "x10", "10x", ""],
    },
    {
        "type": BaseFilter.PositiveNonZeroNumber,
        "values": [("1", 1), ("10", 10), ("99", 99), (" 1 ", 1), ("010", 10)],
        "invalid_values": ["-1", "x", "x10", "10x", "0", ""],
    },
    {
        "type": BaseFilter.NegativeNumber,
        "values": [("-1", -1), ("-10", -10), (" -9 ", -9), ("-0", 0)],
        "invalid_values": ["x", "x10", "10x", "- 1", ""],
    },
    {
        "type": BaseFilter.DecimalNumber,
        "values": [
            ("1", 1.0),
            ("10", 10.0),
            ("99", 99.0),
            ("-1", -1.0),
            ("-10", -10.0),
            ("010", 10.0),
            ("  1 ", 1.0),
            ("1.0", 1.0),
            ("10.12", 10.12),
            ("9.9", 9.9),
            ("-1.1", -1.1),
            (" -10.2 ", -10.2),
            ("  1 ", 1.0),
            (".11", 0.11),
            ("0.111", 0.111),
            ("0", 0.0),
        ],
        "invalid_values": ["x", "x10", "10x", "- 1.1", "", "."],
    },
    {
        "type": BaseFilter.String,
        "values": [
            ("a", "a"),
            ("bbbb", "bbbb"),
            ("  cccc  ", "cccc"),
            ("  cc:cc  ", "cc:cc"),
            ("'a,b'", "a,b"),
        ],
        "invalid_values": ["", ",", ",,,,"],
    },
    {
        "type": BaseFilter.Boolean,
        "values": [
            ("1", True),
            ("True", True),
            ("true", True),
            ("0", False),
            ("False", False),
            ("false", False),
            (" True ", True),
        ],
        "invalid_values": ["", "x", "TRUE", "111"],
    },
    {
        "type": r"\dx\d",
        "values": [("1x1", "1x1"), (" 9x9   ", "9x9")],
        "invalid_values": ["a", ",", "9 x 9"],
    },
]


class FilterParamsTestCase(TestCase):
    def test_with_valid_values_should_correctly_parse_value(self):
        for params in FILTER_PARAMS_DATA:
            for test_data, expected_data in params["values"]:
                BaseFilter.compile_regex(
                    {"name": "x", "params": [params["type"]]}
                )
                filter_instance = BaseFilter(f"x({test_data})")
                expect(filter_instance.params[0]).to_equal(expected_data)

    def test_with_invalid_values_should_correctly_parse_value(self):
        for params in FILTER_PARAMS_DATA:
            for test_data in params["invalid_values"]:
                BaseFilter.compile_regex(
                    {"name": "x", "params": [params["type"]]}
                )
                filter_instance = BaseFilter(f"x({test_data})")
                expect(filter_instance.params).to_be_null()


class MyFilter(BaseFilter):
    @filter_method(BaseFilter.Number, BaseFilter.DecimalNumber)
    async def my_filter(self, value1, value2):
        return (value1, value2)


class StringFilter(BaseFilter):
    @filter_method(BaseFilter.String)
    async def my_string_filter(self, value):
        return value


class EmptyFilter(BaseFilter):
    @filter_method()
    async def my_empty_filter(self):
        return "ok"


class InvalidFilter(BaseFilter):
    async def my_invalid_filter(self, value):
        return value


class DoubleStringFilter(BaseFilter):
    @filter_method(BaseFilter.String, BaseFilter.String)
    async def my_string_filter(self, value1, value2):
        return (value1, value2)


class OptionalParamFilter(BaseFilter):
    @filter_method(BaseFilter.String, BaseFilter.String)
    async def my_optional_filter(self, value1, value2="not provided"):
        return (value1, value2)


class PreLoadFilter(BaseFilter):
    phase = thumbor.filters.PHASE_PRE_LOAD

    @filter_method(BaseFilter.String)
    async def my_pre_load_filter(self, value):
        return value


class BaseFilterTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.context = self.get_context()
        self.factory = FiltersFactory(
            [MyFilter, StringFilter, OptionalParamFilter, PreLoadFilter]
        )
        self.runner = self.get_runner()

    def get_runner(self):
        return None

    def get_context(self):
        def is_multiple():
            return False

        cfg = Config()
        importer = Importer(cfg)
        importer.import_modules()
        context = Context(config=cfg, importer=importer)
        context.modules.engine.is_multiple = is_multiple
        return context


class RunnerWithParametersFilterTestCase(BaseFilterTestCase):
    def get_runner(self):
        return self.factory.create_instances(
            self.context,
            "my_string_filter(aaaa):my_string_filter(bbb):my_pre_load_filter(ccc)",
        )

    def test_runner_with_parameters_should_create_two_instances(self):
        post_instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        pre_instances = self.runner.filter_instances[
            thumbor.filters.PHASE_PRE_LOAD
        ]
        expect(len(post_instances)).to_equal(2)
        expect(post_instances[0].__class__).to_equal(StringFilter)
        expect(post_instances[1].__class__).to_equal(StringFilter)
        expect(len(pre_instances)).to_equal(1)
        expect(pre_instances[0].__class__).to_equal(PreLoadFilter)

    @gen_test
    async def test_running_post_filters_should_run_only_post_filters(self):
        await self.runner.apply_filters(thumbor.filters.PHASE_POST_TRANSFORM)
        post_instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        pre_instances = self.runner.filter_instances[
            thumbor.filters.PHASE_PRE_LOAD
        ]
        expect(len(post_instances)).to_equal(0)
        expect(len(pre_instances)).to_equal(1)

    @gen_test
    async def test_running_pre_filters_should_run_only_pre_filters(self):
        await self.runner.apply_filters(thumbor.filters.PHASE_POST_TRANSFORM)
        await self.runner.apply_filters(thumbor.filters.PHASE_PRE_LOAD)
        post_instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        pre_instances = self.runner.filter_instances[
            thumbor.filters.PHASE_PRE_LOAD
        ]
        expect(len(post_instances)).to_equal(0)
        expect(len(pre_instances)).to_equal(0)

    def test_invalid_filter(self):
        InvalidFilter.pre_compile()
        expect(hasattr(InvalidFilter, "runnable_method")).to_be_false()

    def test_valid_filter_creates_a_runnable_method(self):
        MyFilter.pre_compile()
        expect(MyFilter.runnable_method).to_equal(MyFilter.my_filter)

    @gen_test
    async def test_valid_filter_sets_correct_result_value(self):
        filter_instance = MyFilter("my_filter(1, -1.1)")
        result = await filter_instance.run()
        expect(result).to_equal([(1, -1.1)])

    @gen_test
    async def test_invalid_number_throws_an_error(self):
        filter_instance = MyFilter("my_invalid_filter(x, 1)")
        result = await filter_instance.run()
        expect(hasattr(result, "result")).to_be_false()

    @gen_test
    async def test_double_string_filter_sets_correct_values(self):
        DoubleStringFilter.pre_compile()
        filter_instance = DoubleStringFilter("my_string_filter(a, b)")
        result = await filter_instance.run()
        expect(result).to_equal([("a", "b")])

    @gen_test
    async def test_with_strings_with_commas_sets_correct_values(self):
        DoubleStringFilter.pre_compile()
        tests = [
            ("my_string_filter(a,'b, c')", [("a", "b, c")]),
            ("my_string_filter('a,b', c)", [("a,b", "c")]),
            ("my_string_filter('ab', c)", [("ab", "c")]),
            ("my_string_filter('ab,', c)", [("ab,", "c")]),
            ("my_string_filter('ab,', ',c')", [("ab,", ",c")]),
            ("my_string_filter('ab, c)", [("'ab", "c")]),
            ("my_string_filter('ab, c',d)", [("ab, c", "d")]),
            ("my_string_filter('a,b, c)", None),
            ("my_string_filter('a,b, c')", None),
        ]
        for test, expected in tests:
            filter_instance = DoubleStringFilter(test)
            result = await filter_instance.run()
            expect(result).to_equal(expected)

    @gen_test
    async def test_with_empty_filter_should_call_filter(self):
        EmptyFilter.pre_compile()
        filter_instance = EmptyFilter("my_empty_filter()")
        result = await filter_instance.run()
        expect(result).to_equal(["ok"])

    @pytest.mark.skipif(
        sys.version_info < (3, 8),
        reason="python37 mock does not support async mocks, see https://bugs.python.org/issue26467",
    )
    @gen_test
    @patch.object(StringFilter, "run", autospec=True)
    async def test_apply_filters_respects_filter_params_order(self, run_mock):
        filter_order = []
        run_mock.side_effect = lambda self: filter_order.append(self.params[0])
        await self.runner.apply_filters(thumbor.filters.PHASE_POST_TRANSFORM)
        assert filter_order == ["aaaa", "bbb"]


class WithOneValidParamFilterTestCase(BaseFilterTestCase):
    def get_runner(self):
        return self.factory.create_instances(
            self.context, "my_filter(1, 0a):my_string_filter(aaaa)"
        )

    def test_should_create_one_instance(self):
        instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        expect(len(instances)).to_equal(1)
        expect(instances[0].__class__).to_equal(StringFilter)


class WithParameterContainingColonsFilterTestCase(BaseFilterTestCase):
    def get_runner(self):
        return self.factory.create_instances(
            self.context, "my_string_filter(aaaa):my_string_filter(aa:aa)"
        )

    def test_should_create_two_instances(self):
        instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        expect(len(instances)).to_equal(2)
        expect(instances[0].__class__).to_equal(StringFilter)
        expect(instances[1].__class__).to_equal(StringFilter)

    def test_should_understant_parameters(self):
        instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        expect(instances[0].params).to_equal(["aaaa"])
        expect(instances[1].params).to_equal(["aa:aa"])


class WithValidParamsFilterTestCase(BaseFilterTestCase):
    def get_runner(self):
        return self.factory.create_instances(
            self.context, "my_filter(1, 0):my_string_filter(aaaa)"
        )

    def test_should_create_two_instances(self):
        instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        expect(len(instances)).to_equal(2)
        expect(instances[0].__class__).to_equal(MyFilter)
        expect(instances[1].__class__).to_equal(StringFilter)

    @gen_test
    async def test_when_running_should_create_two_instances(self):
        result = []
        instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        for instance in instances:
            result.append(await instance.run())
        expect(result[0]).to_equal([(1, 0.0)])
        expect(result[1]).to_equal(["aaaa"])


class WithOptionalParamFilterTestCase(BaseFilterTestCase):
    def get_runner(self):
        return self.factory.create_instances(
            self.context, "my_optional_filter(aa, bb)"
        )

    def test_should_create_two_instances(self):
        instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        expect(len(instances)).to_equal(1)
        expect(instances[0].__class__).to_equal(OptionalParamFilter)

    @gen_test
    async def test_should_understand_parameters(self):
        instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        expect(await instances[0].run()).to_equal([("aa", "bb")])


class WithOptionalParamsInOptionalFilterTestCase(BaseFilterTestCase):
    def get_runner(self):
        return self.factory.create_instances(
            self.context, "my_optional_filter(aa)"
        )

    def test_should_create_two_instances(self):
        instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        expect(len(instances)).to_equal(1)
        expect(instances[0].__class__).to_equal(OptionalParamFilter)

    @gen_test
    async def test_should_understand_parameters(self):
        instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        result = await instances[0].run()
        expect(result).to_equal([("aa", "not provided")])


class WithInvalidOptionalFilterTestCase(BaseFilterTestCase):
    def get_runner(self):
        return self.factory.create_instances(
            self.context, "my_optional_filter()"
        )

    def test_should_create_two_instances(self):
        instances = self.runner.filter_instances[
            thumbor.filters.PHASE_POST_TRANSFORM
        ]
        expect(len(instances)).to_equal(0)


class WithPreLoadFilterTestCase(BaseFilterTestCase):
    def get_runner(self):
        return self.factory.create_instances(
            self.context, "my_pre_load_filter(aaaa)"
        )

    def should_create_two_instances(self):
        instances = self.runner.filter_instances[
            thumbor.filters.PHASE_PRE_LOAD
        ]
        expect(len(instances)).to_equal(1)
        expect(instances[0].__class__).to_equal(PreLoadFilter)

    def should_understant_parameters(self):
        instances = self.runner.filter_instances[
            thumbor.filters.PHASE_PRE_LOAD
        ]
        expect(instances[0].params).to_equal(["aaaa"])



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import mimetypes
import random
from io import BytesIO
from os.path import dirname, join, realpath
from unittest import mock
from urllib.parse import urlencode

from PIL import Image
from ssim import compute_ssim
from tornado.testing import AsyncHTTPTestCase

from thumbor.app import ThumborServiceApp
from thumbor.config import Config
from thumbor.context import Context, RequestParameters
from thumbor.engines.pil import Engine as PilEngine
from thumbor.importer import Importer
from thumbor.transformer import Transformer


def get_ssim(actual, expected):
    if (
        actual.size[0] != expected.size[0]
        or actual.size[1] != expected.size[1]
    ):
        raise RuntimeError(
            "Can't calculate SSIM for images of different sizes ("
            f"one is {actual.size[0]}x{actual.size[1]}, "
            f"the other {expected.size[0]}x{expected.size[1]}).",
        )

    return compute_ssim(actual, expected)


def encode_multipart_formdata(fields, files):
    boundary = b"thumborUploadFormBoundary"
    crlf = b"\r\n"
    lines = []
    for key, value in fields.items():
        lines.append(b"--" + boundary)
        lines.append(
            b'Content-Disposition: form-data; name="%s"' % key.encode()
        )
        lines.append(b"")
        lines.append(value)
    for (key, filename, value) in files:
        lines.append(b"--" + boundary)
        lines.append(
            b'Content-Disposition: form-data; name="%s"; filename="%s"'
            % (key.encode(), filename.encode())
        )
        lines.append(
            b"Content-Type: %s" % mimetypes.guess_type(filename)[0].encode()
            or b"application/octet-stream"
        )
        lines.append(b"")
        lines.append(value)
    lines.append(b"")
    lines.append(b"")
    lines.append(b"--" + boundary + b"--")
    body = crlf.join(lines)
    content_type = b"multipart/form-data; boundary=%s" % boundary
    return content_type, body


class TestCase(AsyncHTTPTestCase):
    _multiprocess_can_split_ = True

    def get_app(self):
        self.config = (  # This is a test case pylint: disable=attribute-defined-outside-init
            self.get_config()
        )
        self.server = (  # This is a test case pylint: disable=attribute-defined-outside-init
            self.get_server()
        )
        self.importer = (  # This is a test case pylint: disable=attribute-defined-outside-init
            self.get_importer()
        )
        self.request_handler = (  # This is a test case pylint: disable=attribute-defined-outside-init
            self.get_request_handler()
        )
        self.importer.import_modules()

        self.context = self.get_context()
        return ThumborServiceApp(self.context)

    def get_config(self):  # Meant to be overriden pylint: disable=no-self-use
        return Config()

    def get_server(self):  # Meant to be overriden pylint: disable=no-self-use
        return None

    def get_importer(self):
        importer = Importer(self.config)
        importer.import_modules()
        return importer

    def get_request_handler(
        self,
    ):  # Meant to be overriden pylint: disable=no-self-use
        return None

    def get_context(self):
        self.importer.import_modules()
        return Context(
            self.server, self.config, self.importer, self.request_handler
        )

    async def async_fetch(self, path, method="GET", body=None, headers=None):
        return await self.http_client.fetch(
            self.get_url(path),
            method=method,
            body=body,
            headers=headers,
            allow_nonstandard_methods=True,
            raise_error=False,
        )

    async def async_get(self, path, headers=None):
        return await self.async_fetch(
            path, method="GET", body=urlencode({}, doseq=True), headers=headers
        )

    async def async_post(self, path, headers, body):
        return await self.async_fetch(
            path,
            method="POST",
            body=body,
            headers=headers,
        )

    async def async_put(self, path, headers, body):
        return await self.async_fetch(
            path,
            method="PUT",
            body=body,
            headers=headers,
        )

    async def async_delete(self, path, headers):
        return await self.async_fetch(
            path,
            method="DELETE",
            body=urlencode({}, doseq=True),
            headers=headers,
        )

    async def async_post_files(self, path, data=None, files=None):
        if data is None:
            data = {}
        if files is None:
            files = []
        multipart_data = encode_multipart_formdata(data, files)

        return await self.async_fetch(
            path,
            method="POST",
            body=multipart_data[1],
            headers={"Content-Type": multipart_data[0]},
        )


class FilterTestCase(TestCase):
    _multiprocess_can_split_ = True

    def setUp(self):
        super().setUp()
        self.context = {}

    def get_filter(self, filter_name, params_string="", config_context=None):
        config = Config(
            FILTERS=[filter_name],
            LOADER="thumbor.loaders.file_loader",
            FILE_LOADER_ROOT_PATH=self.get_fixture_root_path(),
        )
        importer = Importer(config)
        importer.import_modules()

        req = RequestParameters()

        context = Context(config=config, importer=importer)
        context.request = req
        context.request.engine = context.modules.engine

        if config_context is not None:
            config_context(context)

        self.context = context

        fltr = importer.filters[0]
        fltr.pre_compile()

        context.transformer = Transformer(context)

        return fltr(params_string, context=context)

    @staticmethod
    def get_fixture_root_path():
        return join(dirname(realpath(__file__)), "fixtures", "filters")

    def get_fixture_path(self, name):
        return f"{self.get_fixture_root_path()}/{name}"

    def get_fixture(self, name, mode="RGB"):
        image = Image.open(self.get_fixture_path(name))
        return image.convert(mode)

    async def get_filtered(
        self,
        source_image,
        filter_name,
        params_string,
        config_context=None,
        mode="RGB",
    ):
        fltr = self.get_filter(filter_name, params_string, config_context)
        image = Image.open(self.get_fixture_path(source_image))
        img_buffer = BytesIO()

        # Special case for the quality test, because the quality filter
        # doesn't really affect the image, it only sets a context value
        # for use on save. But here we convert the result,
        # we do not save it
        if params_string == "quality(10)":
            image.save(img_buffer, "JPEG", quality=10)
            fltr.engine.load(img_buffer.getvalue(), ".jpg")
        else:
            image.save(img_buffer, "PNG", quality=100)
            fltr.engine.load(img_buffer.getvalue(), ".png")

        fltr.context.transformer.img_operation_worker()

        await fltr.run()

        fltr.engine.image = fltr.engine.image.convert(mode)

        return fltr.engine.image

    @staticmethod
    def get_ssim(actual, expected):
        return get_ssim(actual, expected)

    def debug(self, image):  # pylint: disable=arguments-differ
        image = Image.fromarray(image)
        path = f"/tmp/debug_image_{random.randint(1, 10000)}.jpg"
        image.save(path, "JPEG")
        print(f"The debug image was in {path}.")

    @staticmethod
    def debug_size(image):
        loaded = Image.fromarray(image)
        print(
            f"Image dimensions are {loaded.size[0]}x{loaded.size[1]} (shape is {image.shape})"
        )


class DetectorTestCase(TestCase):
    _multiprocess_can_split_ = True

    def setUp(self):
        super().setUp()
        self.context.request = mock.Mock(focal_points=[])
        self.engine = PilEngine(self.context)
        self.context.modules.engine = self.engine



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import os
from os.path import exists


class BaseStorage:
    def __init__(self, context):
        self.context = context

    async def put(self, path, file_bytes):
        """
        :returns: Nothing. This method is expected to be asynchronous.
        :rtype: None
        """
        raise NotImplementedError()

    async def put_crypto(self, path):
        """
        :returns: Nothing. This method is expected to be asynchronous.
        :rtype: None
        """
        raise NotImplementedError()

    async def put_detector_data(self, path, data):
        """
        :returns: Nothing. This method is expected to be asynchronous.
        :rtype: None
        """
        raise NotImplementedError()

    async def get_crypto(self, path):
        raise NotImplementedError()

    async def get_detector_data(self, path):
        raise NotImplementedError()

    async def get(self, path):
        raise NotImplementedError()

    async def exists(self, path):
        raise NotImplementedError()

    async def remove(self, path):
        raise NotImplementedError()

    def ensure_dir(self, path):
        if not exists(path):
            try:
                os.makedirs(path)
            except OSError as err:
                # FILE ALREADY EXISTS = 17
                if err.errno != 17:
                    raise



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from shutil import which
from urllib.parse import quote

from libthumbor import CryptoURL
from preggy import expect
from tornado.testing import gen_test

from tests.base import normalize_unicode_path
from tests.handlers.test_base_handler import BaseImagingTestCase
from thumbor.config import Config
from thumbor.context import Context, ServerParameters
from thumbor.importer import Importer

# pylint: disable=broad-except,abstract-method,attribute-defined-outside-init,line-too-long,too-many-public-methods
# pylint: disable=too-many-lines

JPEGTRAN_AVAILABLE = which("jpegtran") is not None
EXIFTOOL_AVAILABLE = which("exiftool") is not None


class ImagingOperationsWithHttpLoaderTestCase(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.http_loader"
        cfg.STORAGE = "thumbor.storages.file_storage"
        cfg.FILE_STORAGE_ROOT_PATH = self.root_path

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        return Context(server, cfg, importer)

    @gen_test
    async def test_image_already_generated_by_thumbor(self):
        with open("./tests/fixtures/images/image.jpg", "rb") as fixture:
            await self.context.modules.storage.put(
                quote("http://test.com/smart/image.jpg"), fixture.read()
            )
        crypto = CryptoURL("ACME-SEC")
        image_url = self.get_url(
            crypto.generate(image_url=quote("http://test.com/smart/image.jpg"))
        )
        url = crypto.generate(image_url=quote(image_url))

        response = await self.async_fetch(url)
        expect(response.code).to_equal(200)

    @gen_test
    async def test_image_already_generated_by_thumbor_2_times(self):
        with open(
            normalize_unicode_path(
                "./tests/fixtures/images/alabama1_ap620é.jpg"
            ),
            "rb",
        ) as fixture:
            await self.context.modules.storage.put(
                quote("http://test.com/smart/alabama1_ap620é"), fixture.read()
            )
        crypto = CryptoURL("ACME-SEC")
        image_url = self.get_url(
            crypto.generate(
                image_url=quote(
                    self.get_url(
                        crypto.generate(
                            image_url=quote(
                                "http://test.com/smart/alabama1_ap620é"
                            )
                        )
                    )
                )
            )
        )

        url = crypto.generate(image_url=quote(image_url))

        response = await self.async_fetch(url)
        expect(response.code).to_equal(200)

    @gen_test
    async def test_image_with_utf8_url(self):
        with open("./tests/fixtures/images/maracujá.jpg", "rb") as fixture:
            await self.context.modules.storage.put(
                quote("http://test.com/maracujá.jpg".encode("utf-8")),
                fixture.read(),
            )
        crypto = CryptoURL("ACME-SEC")
        image_url = self.get_url(
            quote("/unsafe/http://test.com/maracujá.jpg".encode("utf-8"))
        )
        url = crypto.generate(image_url=quote(image_url))
        response = await self.async_fetch(url)
        expect(response.code).to_equal(200)

    @gen_test
    async def test_image_with_http_utf8_url(self):
        with open("./tests/fixtures/images/maracujá.jpg", "rb") as fixture:
            await self.context.modules.storage.put(
                quote("http://test.com/maracujá.jpg".encode("utf-8")),
                fixture.read(),
            )

        url = quote("/unsafe/http://test.com/maracujá.jpg".encode("utf-8"))
        response = await self.async_fetch(url)
        expect(response.code).to_equal(200)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import re
from io import BytesIO
from shutil import which
from subprocess import PIPE, Popen

from PIL import Image

from thumbor.engines.pil import Engine as PILEngine
from thumbor.utils import logger

GIFSICLE_SIZE_REGEX = re.compile(r"(?:logical\sscreen\s(\d+x\d+))")
GIFSICLE_IMAGE_COUNT_REGEX = re.compile(r"(?:(\d+)\simage)")


# TODO: Rename this weird error to GifsicleError
class GifSicleError(RuntimeError):
    pass


class Engine(PILEngine):
    @property
    def size(self):
        return self.image_size

    def run_gifsicle(self, command):
        if self.context.server.gifsicle_path is None or not which(
            self.context.server.gifsicle_path
        ):
            raise GifSicleError(
                "gifsicle command was not found and it is required"
                " for your configuration of Thumbor"
            )
        process = Popen(  # pylint: disable=consider-using-with
            [self.context.server.gifsicle_path] + command.split(" "),
            stdout=PIPE,
            stdin=PIPE,
            stderr=PIPE,
        )
        stdout_data, stderr_data = process.communicate(input=self.buffer)
        if process.returncode != 0:
            logger.error(stderr_data)

        if stdout_data is None:
            gifsicle_command = (
                " ".join(
                    [self.context.server.gifsicle_path]
                    + command.split(" ")
                    + [self.context.request.url]
                ),
            )

            raise GifSicleError(
                (
                    f"gifsicle command returned errorlevel {process.returncode} for "
                    f'command "{gifsicle_command}" (image maybe corrupted?)'
                )
            )

        return stdout_data

    def is_multiple(self):
        return self.frame_count > 1

    def update_image_info(self):
        self._is_multiple = False

        result = self.run_gifsicle("--info").decode("utf-8")
        size = GIFSICLE_SIZE_REGEX.search(result)
        self.image_size = size.groups()[0].split("x")
        self.image_size[0], self.image_size[1] = (
            int(self.image_size[0]),
            int(self.image_size[1]),
        )

        count = GIFSICLE_IMAGE_COUNT_REGEX.search(result)
        self.frame_count = int(count.groups()[0])

    def load(self, buffer, extension):
        self.extension = extension
        self.buffer = buffer
        self.image = ""
        self.operations = []
        self.update_image_info()

    def draw_rectangle(self, x, y, width, height):
        raise NotImplementedError()

    def resize(self, width, height):
        if width == 0 and height == 0:
            return

        if width > 0 and height == 0:
            arguments = f"--resize-width {width}"
        elif height > 0 and width == 0:
            arguments = f"--resize-height {height}"
        else:
            arguments = f"--resize {width}x{height}"

        self.operations.append(arguments)

    def crop(self, left, top, right, bottom):
        arguments = f"--crop {left},{top}-{right},{bottom}"
        self.operations.append(arguments)
        self.flush_operations()
        self.update_image_info()

    def rotate(self, degrees):
        if degrees not in [90, 180, 270]:
            return
        arguments = f"--rotate-{degrees}"
        self.operations.append(arguments)

    def flip_vertically(self):
        self.operations.append("--flip-vertical")

    def flip_horizontally(self):
        self.operations.append("--flip-horizontal")

    def extract_cover(self):
        arguments = "#0"
        self.operations.append(arguments)
        self.flush_operations()
        self.update_image_info()

    def flush_operations(self, update_image=True):
        if not self.operations:
            return self.buffer

        buffer = self.run_gifsicle(" ".join(self.operations))

        self.operations = []

        if update_image:
            self.buffer = buffer

        return buffer

    def read(self, extension=None, quality=None):
        return self._read()

    def _read(self, update_image=True):
        buffer = self.flush_operations(update_image)

        # Make sure gifsicle produced a valid gif.
        try:
            with BytesIO(buffer) as buff:
                Image.open(buff).verify()
        except Exception:
            self.context.metrics.incr("gif_engine.no_output")
            logger.error(
                "[GIF_ENGINE] invalid gif engine result for url `%s`.",
                self.context.request.url,
            )
            raise

        return buffer

    def convert_to_grayscale(self, update_image=True, alpha=True):
        self.operations.append("--use-colormap gray")
        return self._read(update_image)

    # gif have no exif data and thus can't be auto oriented
    def reorientate(self, override_exif=True):
        pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from os.path import abspath, dirname, join

import tornado.web
from preggy import expect
from tornado.testing import gen_test

import thumbor.loaders.https_loader as loader
from tests.base import TestCase
from thumbor.config import Config
from thumbor.context import Context
from thumbor.loaders import LoaderResult

# test code
# pylint: disable=abstract-method,too-few-public-methods,protected-access


def fixture_for(filename):
    return abspath(join(dirname(__file__), "fixtures", filename))


class MainHandler(tornado.web.RequestHandler):
    def get(self):
        self.write("Hello")


class EchoUserAgentHandler(tornado.web.RequestHandler):
    def get(self):
        self.write(self.request.headers["User-Agent"])


class HandlerMock:
    def __init__(self, headers):
        self.request = RequestMock(headers)


class RequestMock:
    def __init__(self, headers):
        self.headers = headers


class ResponseMock:
    def __init__(self, error=None, content_type=None, body=None, code=None):
        self.error = error
        self.code = code
        self.time_info = None

        self.headers = {"Content-Type": "image/jpeg"}

        if content_type:
            self.headers["Content-Type"] = content_type

        self.body = body


class ValidateUrlTestCase(TestCase):
    @gen_test
    async def test_with_allowed_sources(self):
        config = Config()
        config.ALLOWED_SOURCES = ["s.glbimg.com"]
        ctx = Context(None, config, None)
        expect(
            loader.validate(ctx, "http://www.google.com/logo.jpg")
        ).to_be_false()
        expect(
            loader.validate(ctx, "http://s2.glbimg.com/logo.jpg")
        ).to_be_false()
        expect(
            loader.validate(
                ctx,
                "/glob=:sfoir%20%20%3Co-pmb%20%20%20%20_%20%20%20%200%20%20g.-%3E%3Ca%20hplass=",  # NOQA, pylint:disable=line-too-long
            )
        ).to_be_false()
        expect(
            loader.validate(ctx, "http://s.glbimg.com/logo.jpg")
        ).to_be_true()

    @gen_test
    async def test_without_allowed_sources(self):
        config = Config()
        config.ALLOWED_SOURCES = []
        ctx = Context(None, config, None)
        is_valid = loader.validate(ctx, "http://www.google.com/logo.jpg")
        expect(is_valid).to_be_true()


class NormalizeUrlTestCase(TestCase):
    @gen_test
    async def test_should_normalize_url(self):
        expect(loader._normalize_url("http://some.url")).to_equal(
            "http://some.url"
        )
        expect(loader._normalize_url("some.url")).to_equal("https://some.url")

    @gen_test
    async def test_should_normalize_quoted_url(self):
        url = (
            "https%3A//www.google.ca/images/branding/googlelogo/2x/"
            "googlelogo_color_272x92dp.png"
        )
        expected = (
            "https://www.google.ca/images/branding/googlelogo/2x/"
            "googlelogo_color_272x92dp.png"
        )
        result = loader._normalize_url(url)
        expect(result).to_equal(expected)


class HttpsLoaderTestCase(TestCase):
    def get_app(self):
        application = tornado.web.Application([(r"/", MainHandler)])

        return application

    @gen_test
    async def test_load(self):
        url = self.get_url("/")
        config = Config()
        ctx = Context(None, config, None)

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_equal("Hello")
        expect(result.successful).to_be_true()

    @gen_test
    async def test_load_with_curl(self):
        url = self.get_url("/")
        config = Config()
        config.HTTP_LOADER_CURL_ASYNC_HTTP_CLIENT = True
        ctx = Context(None, config, None)

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_equal("Hello")
        expect(result.successful).to_be_true()


class HttpLoaderWithUserAgentForwardingTestCase(TestCase):
    def get_app(self):
        application = tornado.web.Application([(r"/", EchoUserAgentHandler)])

        return application

    @gen_test
    async def test_load_with_user_agent(self):
        url = self.get_url("/")
        config = Config()
        config.HTTP_LOADER_FORWARD_USER_AGENT = True
        ctx = Context(
            None, config, None, HandlerMock({"User-Agent": "test-user-agent"})
        )

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_equal("test-user-agent")

    @gen_test
    async def test_load_with_default_user_agent(self):
        url = self.get_url("/")
        config = Config()
        config.HTTP_LOADER_FORWARD_USER_AGENT = True
        config.HTTP_LOADER_DEFAULT_USER_AGENT = "DEFAULT_USER_AGENT"
        ctx = Context(None, config, None, HandlerMock({}))

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_equal("DEFAULT_USER_AGENT")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

# pylint: disable=unsupported-membership-test

from io import BytesIO
from os.path import abspath, dirname, join
from unittest import TestCase, mock

import piexif
from PIL import Image
from preggy import expect
import pytest

from tests.base import (
    skip_unless_avif,
    skip_unless_avif_encoder,
    skip_unless_heif,
)
from thumbor.config import Config
from thumbor.context import Context
from thumbor.engines.pil import Engine, KEEP_EXIF_COPYRIGHT_TAGS


FIXTURES_PATH = abspath(join(dirname(__file__), "../fixtures/"))
STORAGE_PATH = join(FIXTURES_PATH, "images")


class PilEngineTestCase(TestCase):
    def get_context(self):
        cfg = Config(
            SECURITY_KEY="ACME-SEC",
            ENGINE="thumbor.engines.pil",
            IMAGE_METADATA_READ_FORMATS="exif,xmp",
        )
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = STORAGE_PATH
        cfg.STORAGE = "thumbor.storages.no_storage"

        return Context(config=cfg)

    def setUp(self):
        self.context = self.get_context()

    def test_create_engine(self):
        engine = Engine(self.context)
        expect(engine).to_be_instance_of(Engine)

    def test_load_image(self):
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "image.jpg"), "rb") as image_file:
            buffer = image_file.read()
        image = engine.create_image(buffer)
        expect(image.format).to_equal("JPEG")
        expect(engine.subsampling).to_equal(0)

    def test_load_jp2_image(self):
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "image.jp2"), "rb") as image_file:
            buffer = image_file.read()
        image = engine.create_image(buffer)
        expect(image.format).to_equal("JPEG2000")
        expect(engine.subsampling).to_be_null()

    def test_load_psd_image(self):
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "1x1.psd"), "rb") as image_file:
            buffer = image_file.read()
        image = engine.create_image(buffer)
        expect(image.format).to_equal("PSD")
        expect(engine.subsampling).to_be_null()

    def test_load_tif_16bit_per_channel_lsb(self):
        engine = Engine(self.context)
        with open(
            join(STORAGE_PATH, "gradient_lsb_16bperchannel.tif"), "rb"
        ) as image_file:
            buffer = image_file.read()
        expect(buffer).not_to_equal(None)
        engine.load(buffer, None)

        final_bytes = BytesIO(engine.read())
        image_file = Image.open(final_bytes)
        expect(image_file.format).to_equal("PNG")
        expect(image_file.size).to_equal((100, 100))

    def test_load_tif_16bit_per_channel_msb(self):
        engine = Engine(self.context)
        with open(
            join(STORAGE_PATH, "gradient_msb_16bperchannel.tif"), "rb"
        ) as image_file:
            buffer = image_file.read()
        expect(buffer).not_to_equal(None)
        engine.load(buffer, None)

        final_bytes = BytesIO(engine.read())
        image_file = Image.open(final_bytes)
        expect(image_file.format).to_equal("PNG")
        expect(image_file.size).to_equal((100, 100))

    def test_load_tif_8bit_per_channel(self):
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "gradient_8bit.tif"), "rb") as image_file:
            buffer = image_file.read()
        expect(buffer).not_to_equal(None)
        engine.load(buffer, None)

        final_bytes = BytesIO(engine.read())
        image_file = Image.open(final_bytes)
        expect(image_file.format).to_equal("PNG")
        expect(image_file.size).to_equal((100, 100))

    def test_convert_png_1bit_to_png(self):
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "1bit.png"), "rb") as image_file:
            buffer = image_file.read()
        engine.load(buffer, ".png")
        expect(engine.original_mode).to_equal(
            "P"
        )  # Note that this is not a true 1bit image, it's 8bit in black/white.

        engine.resize(10, 10)
        mode, _ = engine.image_data_as_rgb()
        expect(mode).to_equal("RGB")

        final_bytes = BytesIO(engine.read())
        mode = Image.open(final_bytes).mode
        expect(mode).to_equal("P")

    def test_convert_should_preserve_palette_mode(self):
        engine = Engine(self.context)
        with open(
            join(STORAGE_PATH, "256_color_palette.png"), "rb"
        ) as image_file:
            buffer = image_file.read()
        engine.load(buffer, ".png")
        expect(engine.original_mode).to_equal("P")

        engine.resize(10, 10)
        mode, _ = engine.image_data_as_rgb()
        expect(mode).to_equal("RGB")

        final_bytes = BytesIO(engine.read())
        mode = Image.open(final_bytes).mode
        expect(mode).to_equal("P")

    def test_can_set_resampling_filter(self):
        to_test = {
            "LANCZOS": Image.LANCZOS,
            "NEAREST": Image.NEAREST,
            "BiLinear": Image.BILINEAR,
            "bicubic": Image.BICUBIC,
            "garbage": Image.LANCZOS,
        }

        if hasattr(Image, "HAMMING"):
            to_test["HAMMING"] = Image.HAMMING

        for setting, expected in to_test.items():
            cfg = Config(PILLOW_RESAMPLING_FILTER=setting)
            engine = Engine(Context(config=cfg))
            expect(engine.get_resize_filter()).to_equal(expected)

        cfg = Config()
        engine = Engine(Context(config=cfg))
        expect(engine.get_resize_filter()).to_equal(Image.LANCZOS)

    def test_resize_truncated_image(self):
        engine = Engine(self.context)
        with open(
            join(STORAGE_PATH, "BlueSquare_truncated.jpg"), "rb"
        ) as image_file:
            buffer = image_file.read()
        engine.load(buffer, ".jpg")
        engine.resize(10, 10)
        mode, _ = engine.image_data_as_rgb()
        expect(mode).to_equal("RGB")

    def test_load_image_with_metadata(self):
        engine = Engine(self.context)
        with open(
            join(STORAGE_PATH, "Christophe_Henner_-_June_2016.jpg"), "rb"
        ) as image_file:
            buffer = image_file.read()

        engine.load(buffer, None)
        image = engine.image
        expect(image.format).to_equal("JPEG")
        expect(engine.metadata).Not.to_be_null()
        expect(engine.metadata.__class__.__name__).to_equal("dict")

        expect(
            engine.metadata["Exif"][piexif.ExifIFD.LensSerialNumber]
        ).to_equal("0000c139be")

        expect(engine.metadata["0th"][piexif.ImageIFD.Software]).to_equal(
            "Adobe Photoshop Lightroom 4.4 (Macintosh)"
        )

        expect(
            engine.metadata["Exif"][piexif.ExifIFD.DateTimeOriginal]
        ).to_equal("2016:06:23 13:18:05")

    def test_should_preserve_png_transparency(self):
        engine = Engine(self.context)

        with open(
            join(STORAGE_PATH, "paletted-transparent.png"), "rb"
        ) as image_file:
            buffer = image_file.read()

        engine.load(buffer, "png")
        expect(engine.original_mode).to_equal("P")
        engine.resize(200, 150)

        img = Image.open(BytesIO(engine.read(".png")))

        expect(img.mode).to_equal("P")
        expect(img.format.lower()).to_equal("png")

        transparent_pixels_count = sum(
            img.convert("RGBA")
            .split()[3]  # Get alpha channel
            .point(
                lambda x: 0 if x else 1
            )  # return 1 if pixel is transparent, 0 otherwise
            .getdata()
        )

        # Image has total of 200x150=30000 pixels. Most of them should be transparent
        expect(transparent_pixels_count).to_be_greater_than(19000)

    @skip_unless_avif
    def test_convert_jpg_to_avif(self):
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "image.jpg"), "rb") as image_file:
            buffer = image_file.read()
        engine.load(buffer, ".jpg")
        expect(engine.image.format).to_equal("JPEG")
        avif_bytes = BytesIO(engine.read(".avif"))
        with Image.open(avif_bytes) as img:
            expect(img.format).to_equal("AVIF")

    @skip_unless_avif_encoder("aom")
    def test_avif_codec_setting(self):
        self.context.config.AVIF_CODEC = "aom"
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "image.jpg"), "rb") as image_file:
            buffer = image_file.read()
        engine.load(buffer, ".jpg")

        with mock.patch("PIL.ImageFile.ImageFile.save") as mock_save:
            engine.read(".avif")
            mock_save.assert_called_once()
            assert mock_save.call_args[1].get("codec") == "aom"

    @skip_unless_avif
    def test_avif_speed_setting(self):
        self.context.config.AVIF_SPEED = 5
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "image.jpg"), "rb") as image_file:
            buffer = image_file.read()
        engine.load(buffer, ".jpg")

        with mock.patch("PIL.ImageFile.ImageFile.save") as mock_save:
            engine.read(".avif")
            mock_save.assert_called_once()
            assert mock_save.call_args[1].get("speed") == 5

    @skip_unless_avif_encoder("svt")
    def test_avif_svt_odd_dimensions(self):
        self.context.config.AVIF_CODEC = "svt"
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "1bit.png"), "rb") as image_file:
            buffer = image_file.read()
        engine.load(buffer, ".png")
        expect(engine.size).to_equal((691, 212))
        avif_bytes = BytesIO(engine.read(".avif"))
        with Image.open(avif_bytes) as img:
            expect(img.format).to_equal("AVIF")
            expect(img.size).to_equal((690, 212))

    @skip_unless_heif
    def test_convert_jpg_to_heif(self):
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "image.jpg"), "rb") as image_file:
            buffer = image_file.read()
        engine.load(buffer, ".jpg")
        expect(engine.image.format).to_equal("JPEG")
        heif_bytes = BytesIO(engine.read(".heic"))
        with Image.open(heif_bytes) as img:
            expect(img.format).to_equal("HEIF")

    @skip_unless_heif
    def test_heif_odd_dimensions(self):
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "1bit.png"), "rb") as image_file:
            buffer = image_file.read()
        engine.load(buffer, ".png")
        expect(engine.size).to_equal((691, 212))
        heif_bytes = BytesIO(engine.read(".heic"))
        with Image.open(heif_bytes) as img:
            expect(img.format).to_equal("HEIF")
            expect(img.size).to_equal((691, 212))

    def test_should_preserve_copyright_exif_in_image(self):
        self.context.config.PRESERVE_EXIF_COPYRIGHT_INFO = True
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "thumbor-exif.png"), "rb") as image_file:
            buffer_image = image_file.read()
        engine.load(buffer_image, None)

        final_bytes = BytesIO(engine.read())
        image = Image.open(final_bytes)
        expect(image.info.get("exif")).not_to_be_null()

        exifs = piexif.load(image.info.get("exif"))
        expect(exifs["0th"]).not_to_be_null()
        expect(len(exifs["0th"].items())).to_equal(3)
        for k in KEEP_EXIF_COPYRIGHT_TAGS:
            expect(exifs["0th"][k]).not_to_be_null()

    def test_should_read_image_without_copyright_exif(self):
        engine = Engine(self.context)
        with open(join(STORAGE_PATH, "1bit.png"), "rb") as image_file:
            buffer = image_file.read()
        engine.load(buffer, None)

        final_bytes = BytesIO(engine.read())
        image = Image.open(final_bytes)
        expect(image.info.get("exif")).to_be_null()


@skip_unless_avif_encoder("svt")
@pytest.mark.parametrize(
    "image_basename", ("20x20.jpg", "Christophe_Henner_-_June_2016.jpg")
)
def test_avif_svt_codec_fallback(image_basename):
    context = PilEngineTestCase().get_context()
    context.config.AVIF_CODEC = "svt"
    engine = Engine(context)
    with open(join(STORAGE_PATH, image_basename), "rb") as image_file:
        buffer = image_file.read()
    engine.load(buffer, ".jpg")

    with mock.patch("PIL.ImageFile.ImageFile.save") as mock_save:
        engine.read(".avif")
        mock_save.assert_called_once()
        assert mock_save.call_args[1].get("codec") == "auto"


@pytest.mark.xfail(reason="flaky test depending on the state of a dependency")
@skip_unless_avif
@pytest.mark.parametrize("srgb_profile", [None, "srgb.icc"])
def test_avif_convert_cmyk_color_profile_to_srgb(srgb_profile):
    context = PilEngineTestCase().get_context()
    if srgb_profile:
        context.config.SRGB_PROFILE = join(FIXTURES_PATH, srgb_profile)
    engine = Engine(context)
    with open(join(STORAGE_PATH, "cmyk-icc.jpg"), "rb") as image_file:
        buffer = image_file.read()
    engine.load(buffer, ".jpg")
    expect(engine.image.getpixel((0, 0))).to_equal((102, 1, 45, 0))
    avif_bytes = BytesIO(engine.read(".avif"))
    with Image.open(avif_bytes) as img:
        expect(img.format).to_equal("AVIF")
        expect(img.mode).to_equal("RGB")
        rgb_color = img.getpixel((0, 0))
        expect(rgb_color).to_equal((148, 212, 212))


@skip_unless_avif
def test_avif_convert_gray_color_profile_to_srgb():
    context = PilEngineTestCase().get_context()
    engine = Engine(context)
    with open(join(STORAGE_PATH, "grayscale-icc.jpg"), "rb") as image_file:
        buffer = image_file.read()
    engine.load(buffer, ".jpg")
    expect(engine.image.getpixel((0, 0))).to_equal(186)
    avif_bytes = BytesIO(engine.read(".avif"))
    with Image.open(avif_bytes) as img:
        expect(img.format).to_equal("AVIF")
        expect(img.mode).to_equal("RGB")
        rgb_color = img.getpixel((0, 0))
        expect(rgb_color).to_equal((200, 200, 200))
